//
// Created by cr on 24.08.22.
//

#include "FWIRhsAssemblerLinearized.hpp"

DGVAcousticRHSAssemblerLinearized::DGVAcousticRHSAssemblerLinearized(const Vectors &U,
                                                                     std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
                                                                     const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                                                                     const Vector &materialPoint_, const Vector &m_test,
                                                                     double dt, bool discrete,
                                                                     PMLHandler *handler)
        : DGVAcousticRHSAssembler(), U_m(U), PMLH(handler) {
    Config::Get("rkorder", rkorder);
    if (!(rkorder == -102 || rkorder == -302 || rkorder == -2 || rkorder == -1002 || rkorder == 1002 ||
          rkorder == -2002)) {
        Exit("make sure this works for your chosen rkorder");
    }
    FWILinearizedAssembler linearizedAssembler(problem_fwd_, eleCache_, materialPoint_);
    Matrix Mprime(U_m[0]);
    Matrix Dprime(U_m[0]);

    linearizedAssembler.assembleLinearizedMGalerkin(m_test, Mprime);
    if (problem_fwd_->getNumL() > 0)linearizedAssembler.assembleDGalerkin2(m_test, Dprime);
    else {
        Dprime = 0.0;
    }
    Mprime *= 1.0 / dt;
    if (discrete) {
        Matrix APrime(U_m[0]);
        APrime = 0.0;
        Vector lel(m_test);
        lel.Accumulate();
        linearizedAssembler.assembleLinearizedFluxMatrix(lel, APrime);
        Dprime += (-1.0) * APrime;
    }
    Dprime *= -0.5;
    A = std::make_unique<Matrix>(Dprime);
    B = std::make_unique<Matrix>(Dprime);
    (*A) += (-1.0) * Mprime;
    (*B) += (1.0) * Mprime;
    A->CreateSparse();
    B->CreateSparse();
    if (handler) {
        if (handler->isPML()) {
            VecTmp = std::make_unique<Vector>(U[0]);
            VecTmp->Clear();
        }
    }
}

void DGVAcousticRHSAssemblerLinearized::RHS(double t, Vector &b) const {
    if (PMLH->isPML()) {
        *VecTmp = (*A) * U_m[step + 1] + (*B) * U_m[step];
        b.Clear();
        PMLWaveVectorFWI::transfer(PMLH->GetNonPMLToPML(), *VecTmp, b);
    } else {
        b = (*A) * U_m[step + 1] + (*B) * U_m[step];
    }
    if (rkorder == -102 || rkorder == -302) {
        if (increaseStep) {
            step++;
            increaseStep = false;
        } else {
            increaseStep = true;
        }
    } else if (rkorder == -2 || rkorder == -1002 || rkorder == 1002 || rkorder == -2002) {
        step++;
    } else {
        Exit("check this for this rkorder")
    }
}
