//
// Created by cr on 15.10.21.
//

#include <fstream>
#include "FWIHelpers.hpp"
#include "LinearSolver.hpp"
#include "MeshesCreator.hpp"



std::shared_ptr<RectangularGeometry2D> getSharedGeom(const RectangularConfig &conf) {
    return std::make_shared<RectangularGeometry2D>(std::pair<double, double>({conf.LeftBnd, conf.RightBnd}),
                                                   std::pair<double, double>({conf.BottomBnd, conf.TopBnd}),
                                                   conf.dimX, conf.dimY, conf.boundaries(), conf.name);
}

std::shared_ptr<Meshes> getFWIMesh() {
    string mesh = "";
    Config::Get("Mesh", mesh);
    if (mesh == "Marmousi1") {
        RectangularConfig conf;
        conf.LeftBnd = 0.0;
        conf.RightBnd = 9200.0;
        conf.BottomBnd = 0.0;
        conf.TopBnd = 3000.0;
        conf.dimX = 93;
        conf.dimY = 31;
        conf.bndTop = 1;
        conf.bndBottom = 3;
        conf.bndLeft = 3;
        conf.bndRight = 3;
        conf.name = mesh;
        return std::move(MeshesCreator("RectangularFWI").WithCoarseGeometry(getSharedGeom(conf)).CreateShared());
    } else if (mesh == "Camembert") {
        RectangularConfig conf;
        conf.LeftBnd = 0.0;
        conf.RightBnd = 4800.0;
        conf.BottomBnd = 0.0;
        conf.TopBnd = 6000.0;
        conf.dimX = 65;
        conf.dimY = 81;
        conf.bndTop = 2;
        conf.bndBottom = 2;
        conf.bndLeft = 2;
        conf.bndRight = 2;
        conf.name = mesh;
        return std::move(MeshesCreator("RectangularFWI").WithCoarseGeometry(getSharedGeom(conf)).CreateShared());
    } else if (mesh == "Custom") {
        RectangularConfig conf;
        Config::Get("LeftBnd", conf.LeftBnd);
        Config::Get("RightBnd", conf.RightBnd);
        Config::Get("BottomBnd", conf.BottomBnd);
        Config::Get("TopBnd", conf.TopBnd);
        Config::Get("dimX", conf.dimX);
        Config::Get("dimY", conf.dimY);
        int bnd = 0;
        Config::Get("bnd", bnd);
        conf.bndTop = bnd;
        conf.bndBottom = bnd;
        conf.bndLeft = bnd;
        conf.bndRight = bnd;
        Config::Get("bndTop", conf.bndTop);
        Config::Get("bndBottom", conf.bndBottom);
        Config::Get("bndLeft", conf.bndLeft);
        Config::Get("bndRight", conf.bndRight);

        conf.name = mesh;
        return std::move(MeshesCreator("RectangularFWI").WithCoarseGeometry(getSharedGeom(conf)).CreateShared());

    } else {
        return std::move(MeshesCreator().CreateShared());
    }

}

void fillQuadrature(RVector &c, RVector &b, int order) {
    if (order == 1) {
        c.resize(1);
        c[0] = 0;
        b.resize(1);
        b[0] = 1;
    } else if (order == -1) {
        c.resize(1);
        c[0] = 1;
        b.resize(1);
    b[0] = 1;
  } else if (order == -2) {
      c.resize(1);
      c[0] = 0.5;
      b.resize(1);
      b[0] = 1;
  } else if (order == -21 || order == -221 || order == -102 || order == -302) {
      c.resize(2);
      c[0] = 0;
      c[1] = 1;
      b.resize(2);
      b[0] = 0.5;
      b[1] = 0.5;
  } else {
      Exit("RKorder " + to_string(order) + " Not implemented");
  }
}

bool PointInCellDGTS(const Point &z, const Cell &C) {
//  return PointInCellDGTS_old(z, C);

  static bool initialized = true;
  static bool old = false;
  if (initialized) {
      Config::Get("CountCellsOld", old);
      initialized = false;
  };
  if (old) {
    return PointInCellDGTS_old(z, C);
  } else {
    if (C.dim() == 1) {
      double interval_length = 0.5 * abs(C.Corner(0)[0] - C.Corner(1)[0]);
      return (abs(z[0] - C()[0]) < interval_length + GeometricTolerance);
    } else {
      return wn_PnPoly(z, C);
    }
  }
}

bool PointInCellDGTS_old(const Point &z, const Cell &C) {
  // hardcoded for squares
  Point ce = C();
  Point diff = C.Corner(0) - C.Corner(1);
  int dim = C.dim();
  double edgelength = std::max(abs(diff[0]), abs(diff[1]));
  if (C.dim() == 1) {
    double interval_length = 0.5 * abs(C.Corner(0)[0] - C.Corner(1)[0]);
    return (abs(z[0] - C()[0]) < interval_length + GeometricTolerance);
  }
  if (C.Corners() != 4) {
    Exit("PointInCellDGTS only supports squares!")
  }

  double l1distx = abs(z[0] - ce[0]);
  double l1radius = edgelength * 0.5;
  if (l1distx < l1radius + GeometricTolerance) {
    if (dim == 1) {
      return true;
    }
    double l1disty = abs(z[1] - ce[1]);
    if (l1disty < l1radius + GeometricTolerance) {
      if (dim == 2) {
        return true;
      } else {
          THROW("3D needs to be enabled via flag")
//        double zabs = abs(z[2] - ce[2]);
          double zabs = infty;
        if (zabs < l1radius + GeometricTolerance) {
          return true;
        } else {
          return false;
        }
      }
    } else {
      return false;
    }
  }
  return false;
}

int wn_PnPoly(const Point &P, const Cell &C) {
  int wn = 0;    // the  winding number counter

  // loop through all edges of the polygon
  int n = C.Corners();
  for (int i = 0; i < n; i++) {   // edge from V[i] to  V[i+1]
    int j = (i < n - 1) ? i + 1 : 0;
    const Point &corner_i = C.Corner(i);
    const Point &corner_j = C.Corner(j);
    if (corner_i[1] <= P[1]) {          // start y <= P[1]
      if (corner_j[1] > P[1])      // an upward crossing
        if (isLeft(corner_i, corner_j, P) > 0)  // P left of  edge
          ++wn;            // have  a valid up intersect
    } else {                        // start y > P[1] (no test needed)
      if (corner_j[1] <= P[1])     // a downward crossing
        if (isLeft(corner_i, corner_j, P) < 0)  // P right of  edge
          --wn;            // have  a valid down intersect
    }
  }
  return wn;
}

int cn_PnPoly(const Point &P, const Cell &C) {
    int cn = 0;    // the  crossing number counter

    // loop through all edges of the polygon
    int n = C.Corners();
    for (int i = 0; i < n; i++) {    // edge from corner_i  to corner_j
        int j = (i < n - 1) ? i + 1 : 0;
        const Point &corner_i = C.Corner(i);
        const Point &corner_j = C.Corner(j);
        if (((corner_i[1] <= P[1]) && (corner_j[1] > P[1]))     // an upward crossing
            || ((corner_i[1] > P[1]) && (corner_j[1] <= P[1]))) { // a downward crossing
            // compute  the actual edge-ray intersect x-coordinate
            double vt = (P[1] - corner_i[1]) / (corner_j[1] - corner_i[1]);
            if (P[0] < corner_i[0] + vt * (corner_j[0] - corner_i[0])) // P[0] < intersect
                ++cn;   // a valid crossing of y=P[1] right of P[0]
        }
    }
    return (cn & 1);    // 0 if even (out), and 1 if  odd (in)

}

int wn_PnPoly(const Point &P, const cell &c) {
    return wn_PnPoly(P, *c);
}

bool PointInCellDGTS(const Point &z, const cell &c) {
    return PointInCellDGTS(z, *c);
}


class LinearSolver2 : public LinearSolver {
protected:
    void solve(const Operator &A, const Operator &B, Vector &u, Vector &r, double d0, double epsilon) const override {
        u = B * r;
    }

    void solve(const Operator &A, const Operator &B, Vectors &u, Vectors &r, double d0, double epsilon) const override {
        THROW("Not Implemented!");
    }

public:
    explicit LinearSolver2(Preconditioner *preCond, const std::string &prefix = "Linear") : LinearSolver(preCond,
                                                                                                         prefix) {
    }

    ~LinearSolver2() override {}

    std::string Name() const override { return "LinearSolver"; };
};


auto GetFWITI(int rkorder) -> std::unique_ptr<TimeIntegrator> {
    std::string precon = "PointBlockJacobi";
    Config::Get("Preconditioner_S", precon);
    std::string solver = "LS";
    Config::Get("LinearSolver", solver);
    LinearSolver *LinSolv = nullptr;
    if (rkorder > 0 || rkorder == -1002) {
        LinSolv = new LinearSolver2(GetPC("PointBlockJacobi"));
    } else {
        LinSolv = GetLinearSolver(solver, GetPC(precon));
    }
    return TimeIntegratorCreator(static_cast<TIMEINTEGRATOR>(rkorder)).WithLinearSolver(LinSolv).
            WithLinearSolver(GetLinearSolver(solver, GetPC(precon))).
            WithLinearSolver(GetLinearSolver(solver, GetPC(precon))).
            WithLinearSolver(GetLinearSolver(solver, GetPC(precon))).
            CreateUnique();
}

SourceConf::SourceConf() {
    Config::Get("SignalName", SignalName);
    Config::Get("FrequencyFiltering", FrequencyFiltering);
    Config::Get("SignalFreq", SignalFreq);
    Config::Get("f_source", SignalFreq); // Old way to determine Frequency
    Config::Get("SourceInversion", SrcInversion);

  if(SrcInversion){
      Config::Get("SourceInversionKappa", SrcInversionKappa);
      Config::Get("SourceInversionEps", SrcInversionEps);
    }
    if (FrequencyFiltering) {
        Config::Get("FilteringOrder", FilteringOrder);
        Config::Get("MaterialSmoothPasses", MaterialSmoothPasses);
        Config::Get("FilteringType", FilteringType);

      if (FilteringType == "LowPass") {
        std::vector<double> tmp;
        Config::Get("FrequencyLevels", tmp);
        for (double d : tmp){
          FrequencyLevels.push_back({d});
        }
        SignalDelay = 0.5 * 1.0 / (FrequencyFiltering ? FrequencyLevels[0][0] : SignalFreq);
      }else if (FilteringType == "BandPass"){
        string FrequencyFile;
        Config::Get("FrequencyFile",FrequencyFile);
        int StartFrequency = -1;
        Config::Get("StartOfFrequencyLevels", StartFrequency);
        int EndFrequency = -1;
        Config::Get("EndOfFrequencyLevels", EndFrequency);
        int lvl = 0;
        std::ifstream file(FrequencyFile);
        if (!file.good()) {
          Exit("No text-file found at: " + FrequencyFile)
        }
        std::string line;
        vector<vector<double>> FrequencyLevelsTmp;
        while (getline(file, line)) {
          trim(line);
          if (line.length() == 0) {
            continue;
          }
          size_t equalPos = line.find(',');
          if (equalPos == string::npos) {
            continue;
          }
          string x = line.substr(0, equalPos);
          trim(x);
          string y = line.substr(equalPos + 1);
          trim(y);
          FrequencyLevelsTmp.push_back({stod(x), stod(y)});
        }
        for (int i = StartFrequency; i < EndFrequency; ++i){
          FrequencyLevels.push_back(FrequencyLevelsTmp[i]);
        }

      }
     } else {
        FrequencyLevels.push_back({-1});
    }
  Config::Get("FWIPreconRegularization", PreconRegu);
  if (PreconRegu.size() < FrequencyLevels.size()){
    for (int i = 0; i < FrequencyLevels.size()-PreconRegu.size(); ++i){
      PreconRegu.push_back(1e-8);
    }
  }
  Config::Get("SignalCutoff", SignalCutoff);
    Config::Get("SignalDelay", SignalDelay);
    SignalCutoff = 3.0 / SignalFreq + SignalDelay;
    Config::Get("SaveFilteredSeismogram", SaveFilteredSeismogram);
    Config::Get("SaveObservedSeismogram", SaveObservedSeismogram);

}

WPConf::WPConf() {
    Config::Get("numL", numL);
    Config::Get("dim", dim);
    Config::Get("Problem", s_prob);
    Config::Get("level", level);
    RealDataLevel = level;
    Config::Get("RealDataLevel", RealDataLevel);
    InitialIterateLevel = level;
    Config::Get("InitialIterateLevel", InitialIterateLevel);


}
