#ifndef _FWINONLINEARSOLVERS_H_
#define _FWINONLINEARSOLVERS_H_

#include <Newton.hpp>
#include "NonlinearOptions.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"

class FWIForwardModel;

class FWISeismogram;

class FWIDataContainerRange;

class FWIDataContainerDomain;

typedef std::vector<FWIDataContainerRange> FWISubproblem;

typedef std::vector<std::vector<double>> NonlinearOutput;

class FWINonlinearSolver : public NonLinearSolver {
protected:
  int fwiverbose =0;
    Date Start;
    const std::string name;
    FWIForwardModel &FWD;
    NonlinearGeneralOptions generalOpt;
public:
    FWINonlinearSolver(const char *_name, FWIForwardModel &FWD)
            : NonLinearSolver(nullptr, (std::string(_name) + "_").c_str()), FWD(FWD),
              generalOpt(_name), name(_name) {
        d = 0;
        Config::Get("fwiverbose",fwiverbose);
    }

    FWIForwardModel &getFWD() { return FWD; }
    virtual NonlinearOutput operator()(const std::vector<FWIDataContainerRange> &s,
                                       FWIDataContainerDomain &, std::string namesuff) {
        Exit("Not implemented");
    }

    std::string Name() const override { return name; }

    void backtracking(const FWISubproblem &dataObs,
                      std::vector<double> &residualNorms,
                      const FWIDataContainerDomain &iterateMUpdate,
                      FWIDataContainerDomain &iterateNonlinear,
                      FWISubproblem &nonlinearEvaluation,
                      FWISubproblem &defect,
                      std::vector<int> subProb);

    void backtrack(const FWISubproblem &dataObs, std::vector<double> &residualNorms,
                   const FWIDataContainerDomain &iterateMUpdate,
                   FWIDataContainerDomain &iterateNonlinear,
                   FWISubproblem &nonlinearEvaluation,
                   FWISubproblem &defect, std::vector<int> subProb);
    virtual ~FWINonlinearSolver() {}
};

FWINonlinearSolver *GetFWINonlinearSolver(FWIForwardModel &FWD);

#endif
