//#include "m++.hpp"
#include "m++.hpp"
#include "ForwardModels/FWIForwardModelDGTSVA.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "FWINonlinearSolvers.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "VTUModelReader.hpp"
#include "FWIVectorFillFunctions.hpp"
#include "FWIProblemWrapperVA.hpp"

void FWICalcSeismogramsFromVTK() {
  int nrOfModels = 1;
  Config::Get("nrOfModels", nrOfModels);
  std::vector<string> models(nrOfModels, "");
  std::string invParam = "";
  Config::Get("invParam", invParam);
  bool homogenousNames = false;
  Config::Get("homogenousNames", homogenousNames);
  if (homogenousNames) {
    std::string filenameBase ="";
    Config::Get("filenameBase", filenameBase);
    for (int i=0; i<nrOfModels; ++i){
      models[i] = (filenameBase + std::to_string(i)) + ".vtk";
      mout << "Model " << i << ": " << models[i] << endl;
    }
  } else {
    for (int i = 0; i < nrOfModels; ++i) {
      Config::Get(("model" + std::to_string(i)).c_str(), models[i]);
    }
  }
  std::string probConf = "";
  Config::Get("probConf", probConf);
  Settings = probConf.c_str();
  auto meshes = MeshesCreator().CreateShared();
  WPConf Conf;
  LevelPair lev{Conf.level, -1};
  int deg = 0;
  Config::Get("deg", deg);
  FWIForwardModelDGTSVA FWD(meshes, lev, deg, nullptr);
  auto problem = GetInversionProblem(Conf);
  FWD.setProblem(problem);
  problem->setInitialValueFlag(true);
  Vector M(FWD.GetMatVec());
  M.Clear();
//    FWD.fillVectorFromProblem(M, *problem);
  FillVectorFromProblem(FWD.GetProblem(),M);
  std::string measurement_time = "point";
  std::string measurement_space = "point";
  Config::Get("measurement_time", measurement_time);
  Config::Get("measurement_space", measurement_space);
  Vectors U(FWD.TConf.timeSteps, FWD.GetMatVec());
  std::string seismopath = "";
  Exit("VTUMODELREADERHERE!!")
  FWISeismogram s_m;
  int shots = problem->getShotCount();
  for (int i = 0; i < nrOfModels; ++i) {
//        VTKReader.getData(M);
    for (int shot = 0; shot < shots; ++shot) {
//            mout << "Calculating shot " << shot << " with data from " << VTKReader.getCurrentIterateName() << endl;
//            FWD.Parameter2SeismogramMap(M, U, s_m, shot);
      if (shot == 0) FWD.plot(M, ("CorrespondingMaterial" + to_string(i)).c_str());
      s_m.writeToFile("data/FWI/CGREGINN_shot" + to_string(shot) + "_" + to_string(i) + "_full_Residual");
    }
//        VTKReader.nextIterate();
  }
}


void AFFE() {
  size_t shots =6;
  size_t iter = 10;
  std::vector<FWISeismogram> s_obs;
  std::vector<FWISeismogram> s_init;
  std::vector<double> normInit(shots,0.0);
  std::vector<double> initRes(shots,0.0);
  std::vector<double> compRes(iter,0.0);
  std::vector<double> fullRes(iter,0.0);
  std::vector<vector<double>> resNorms;
  std::vector<vector<double>> resFullNorms;
  std::vector<vector<double>> seisNorms;
  std::vector<vector<double>> seisFullNorms;
  double totalInitNorm = 0.0;
  double totalInitRes = 0.0;
  std::string datapath ="";
  Config::Get("datapath", datapath);
  if (datapath == "") Exit("specify data path!")
  std::string name_obs = "CGREGINN_observed_shot";
  std::string name_init = "CGREGINN_initial_shot";
  std::string datapath_seis = datapath + "data/FWI/";
  for(size_t shot =0; shot<shots; ++shot){
    FWISeismogram s;
    std::string fn = datapath_seis + name_init + std::to_string(shot);
    mout << "reading: " << fn << endl;
    s.readFromFile(fn);
    s_init.push_back(s);
    normInit[shot] = s*s;
    totalInitNorm += normInit[shot];
    normInit[shot] = sqrt(normInit[shot]);
  }
  for(size_t shot =0; shot<shots; ++shot){
    FWISeismogram s;
    std::string fn = datapath_seis + name_obs + std::to_string(shot);
    mout << "reading: " << fn << endl;
    s.readFromFile(fn);
    s_obs.push_back(s);
    FWISeismogram s_res(s_init[shot]);
    s_res -= s;
    initRes[shot] = s_res*s_res;
    totalInitRes += initRes[shot];
    initRes[shot] = sqrt(initRes[shot]);
  }
  totalInitRes = sqrt(totalInitRes);
  size_t iters =10;
  std::vector<vector<FWISeismogram>> s_;
  std::string a= "CGREGINN_shot";
  for(size_t iter =0; iter < iters; ++iter){
    std::vector<FWISeismogram> s__;
    std::vector<double> res(shots,0.0);
    std::vector<double> seis(shots,0.0);
    for(size_t shot=0; shot < shots; ++shot){
      std::string  fn = datapath_seis + a + std::to_string(shot) + "_" + std::to_string(iter);
      FWISeismogram s;
      mout << "reading: " << fn << endl;
      s.readFromFile(fn);
      double tmp = s*s;
      seis[shot] = sqrt(tmp);
      FWISeismogram s_res(s_obs[shot]);
      s_res -=s;
      res[shot] = s_res*s_res;
      compRes[iter] += res[shot];
      res[shot] = sqrt(res[shot]);
      s__.push_back(s);
    }
    compRes[iter] = sqrt(compRes[iter]);
    s_.push_back(s__);
    resNorms.push_back(res);
    seisNorms.push_back(seis);
  }
  std::string b = "_full_Residual";
  double totalInitResFull = 0.0;
  std::vector<double> initResFull(shots,0.0);
  for (size_t shot =0; shot<shots; ++shot){
    std::string  fn = datapath_seis  + "NewlyCalculatedSeismos/CGREGINN_initial_shot_" + std::to_string(shot) + b;
    FWISeismogram s;
    mout << "reading: " << fn << endl;
    s.readFromFile(fn);
    initResFull[shot] = s*s;
    totalInitResFull += initResFull[shot];
    initResFull[shot] = sqrt(initResFull[shot]);
  }
  totalInitResFull = sqrt(totalInitResFull);
  for(size_t iter =0; iter < iters; ++iter){
    std::vector<double> seisFull(shots,0.0);
    std::vector<double> resFull(shots,0.0);
    for(size_t shot=0; shot < shots; ++shot){
      std::string  fn = datapath_seis  + "NewlyCalculatedSeismos/"+ a + std::to_string(shot) + "_" + std::to_string(iter)  + b;
      FWISeismogram s;
      mout << "reading: " << fn << endl;
      s.readFromFile(fn);
      double tmp = s*s;
      seisFull[shot] = sqrt(tmp);
      FWISeismogram s_res(s_obs[shot]);
      s_res -= s;
      resFull[shot] = s_res*s_res;
      fullRes[iter] += resFull[shot];
      resFull[shot] = sqrt(resFull[shot]);
    }
    resFullNorms.push_back(resFull);
    fullRes[iter] = sqrt(fullRes[iter]);
    seisFullNorms.push_back(seisFull);
  }

  std::ofstream normFileSeismosRes;
  {
    normFileSeismosRes.open("data/FWI/normFileSeismosRes");
    char SeismoHeaderRes[50];
    std::sprintf(SeismoHeaderRes,"%*s %*s %*s %*s %*s \n",3,"Nr",10,"Comp",10,"Full",10,"QuotComp",10,"QuotFull");
    normFileSeismosRes << SeismoHeaderRes;
    std::sprintf(SeismoHeaderRes,"%*u %*.4f %*.4f",3,999,10,totalInitRes,10,totalInitResFull);
    normFileSeismosRes << SeismoHeaderRes << endl;
  }
  std::ofstream normFileSeismosResFull;
  {
    normFileSeismosResFull.open("data/FWI/normFileSeismosResFull");
    char SeismoHeaderResFull[50];
    std::sprintf(SeismoHeaderResFull,"%*s %*s %*s \n",3,"Nr",10,"Norms",10,"Res");
    normFileSeismosResFull << SeismoHeaderResFull;
  }


  std::ofstream normFileSeismos;
  {
    normFileSeismos.open("data/FWI/normFileSeismos");
    char SeismoHeader[50];
    std::sprintf(SeismoHeader,"%*s %*s %*s\n",3,"Nr",10,"L2",10,"Res");
    normFileSeismos << SeismoHeader;
  }
  std::vector<std::ofstream> normFileSeismosShots(6);
  {
    for(size_t shot =0; shot< normFileSeismosShots.size();++shot){
      normFileSeismosShots[shot].open("data/FWI/normFileSeismos"+std::to_string(shot));
      char SeismoHeaderShots[50];
      std::sprintf(SeismoHeaderShots,"%*s %*s %*s\n",3,"Nr",10,"L2",10,"Res");
      normFileSeismosShots[shot] << SeismoHeaderShots;
    }
  }
  int count =0;
  for(size_t iter=0; iter < iters; ++iter){
    char buff[100];
    for (size_t shot=0; shot < shots;++shot){
      std::sprintf(buff,"%*u %*.4f %*.4f",3,static_cast<int>(iter),10,seisNorms[iter][shot],10,resNorms[iter][shot]);
      normFileSeismosShots[shot] << buff<< endl;

      std::sprintf(buff,"%*u %*.4f %*.4f",3,count,10,seisNorms[iter][shot],10,resNorms[iter][shot]);
      normFileSeismos << buff<< endl;

      std::sprintf(buff,"%*u %*.4f %*.4f",3,count,10,seisFullNorms[iter][shot],10,resFullNorms[iter][shot]);
      normFileSeismosResFull << buff << endl;
      count++;
    }
    double etaComp = 0.0;
    double etaFull = 0.0;
    if (iter==0){
      etaFull = fullRes[iter]/totalInitResFull;
    }else if (iter >0){
      etaComp = compRes[iter]/compRes[iter-1];
      etaFull = fullRes[iter]/fullRes[iter-1];
    }
    std::sprintf(buff,"%*u %*.4f %*.4f %*.4f %*.4f",3,static_cast<int>(iter),10,compRes[iter],10,fullRes[iter],10,etaComp,10,etaFull);
    normFileSeismosRes << buff << endl;
  }
  for(size_t shot=0; shot <normFileSeismosShots.size();++shot){
    normFileSeismosShots[shot].close();
  }
  normFileSeismosResFull.close();
  normFileSeismosRes.close();
  normFileSeismos.close();
}

void FWICreateSeismogramDGTS() {
  auto meshes = getFWIMesh();
  int plevel = 0;
  Config::Get("plevel", plevel);
  int degStart = 0;
  int degEnd = 1;
//    int deg = 2;
  Config::Get("degStart", degStart);
  Config::Get("degEnd", degEnd);
  WPConf Conf;
  for (int deg = degStart; deg <= degEnd; ++deg) {
    for (int scalespace = plevel; scalespace <= Conf.level; ++scalespace) {
      LevelPair lev{scalespace, -1};
      Conf.level = scalespace;
      FWIForwardModelDGTSVA FWD(meshes, lev, deg, GetInversionProblem(Conf));
      Vector M_real(FWD.GetMatVec());
      M_real.Clear();
//            FWD.fillVectorFromProblem(M_real, FWD.GetProblem());
      FillVectorFromProblem(FWD.GetProblem(),M_real);

      FWD.plot(M_real, "fwi_m_real");

      std::string measurement_time = "point";
      std::string measurement_space = "point";
      Config::Get("measurement_time", measurement_time);
      Config::Get("measurement_space", measurement_space);
      int shots = 6;
      int rkorder = -2;
      Config::Get("rkorder", rkorder);
      std::string s_rkorder = "__rkorder_" + std::to_string(rkorder);
      std::string s_deg = std::to_string(deg);
      Config::Get("shots", shots);
      std::string s_level = std::to_string(scalespace);
      double dt_d = 0.01;
      Config::Get("dt", dt_d, -1);
      FWIObservationSpecification *spec = FWIObsSpecWPCreator().Create();
      Vector U_t(FWD.GetWaveVec());
      int Scales = 1;
      Config::Get("scales", Scales);
      double T = 1;
      Config::Get("T", T);
      std::vector<Vector> UU;
      std::vector<double> diffnorms(Scales - 1, 0.0);
      std::vector<double> norms(Scales, 0.0);
      for (int shot = 0; shot < shots; ++shot) {
        FWD.problem_fwd->setShot(shot);
        auto wrap = FWD.GetWrappedProblem(M_real, FWD.GetProblem(), FWD.GetProblem(), true);
        auto RHSDefault = std::make_shared<DGVAcousticRHSAssemblerDefault>(*wrap);
        auto Mat = std::make_unique<DGVAcousticMaterialFromProblem>(*wrap);
        auto matrixAssembler = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), false);
        Config::Get("dt", dt_d, -1);
        for (int scale = 0; scale < Scales; ++scale) {
          if (scale > 0) dt_d *= 0.5;
          MOUT(dt_d);
          auto assembleCombined = std::make_unique<DGVAcousticAssembleCombined>(
              FWD.VAPlotter, matrixAssembler, RHSDefault, 0, T, dt_d);
          assembleCombined->SetEleCache(&FWD.eleCacheContainer->Get());
          Exit("rework for new observation");
//                    if (scale > 0) Psi->refine(0.5);
          FWISeismogram s;
          s.ts_clear();
//                    s.setSpecification(Psi->getObsSpec());
//                    s.setStateComponents(Psi->GetSeismoComps());
          auto timeIntegrator = GetFWITI(rkorder);
          int count = 1;
          U_t.Clear();
          timeIntegrator->Initialize(assembleCombined.get(), U_t);
//                    FWD.measureSingleWaveField(U_t, s, assembleCombined->Time(), assembleCombined->Step());
          timeIntegrator->SetDivergenceCheck(false);
          while (!assembleCombined->IsFinished()) {
            timeIntegrator->OneTimeStep(assembleCombined.get(), U_t);
//                        FWD.measureSingleWaveField(U_t, s, assembleCombined->Time(), assembleCombined->Step());
          }
          timeIntegrator->Destruct();
          UU.push_back(U_t);
          norms[scale] = sqrt(FWD.L2tScalar(U_t, U_t));
          std::string s_shot = std::to_string(shot);
          s.collect();
          s.writeToFile("data/FWI/s__" + measurement_space + "__" + measurement_time + "__level_" +
                        s_level + "__dt_" + std::to_string(dt_d).c_str() + "__deg_" + s_deg + s_rkorder +
                        "__shot_" +
                        s_shot);
        }
        for (int scale = 0; scale < Scales - 1; ++scale) {
          UU[scale] -= UU[scale + 1];
          diffnorms[scale] = sqrt(FWD.L2tScalar(UU[scale], UU[scale]));
        }
        mout << "Norms: ";
        for (int scale = 0; scale < Scales; ++scale) mout << norms[scale] << " ";
        mout << endl;
        mout << "Diffnorms: ";
        for (int scale = 0; scale < Scales - 1; ++scale)
          mout << diffnorms[scale] << " ";
        mout << endl;
        mout << "Quotients: ";
        for (int scale = 1; scale < Scales - 1; ++scale)
          mout << diffnorms[scale - 1] / diffnorms[scale] << " ";
        mout << endl;
      }
    }
  }
}

int FWIConv() {
  std::vector<Vectors> U_ms;
  auto meshes = getFWIMesh();
  WPConf Conf;
  int degStart = 0;
  int degEnd = 1;
  Config::Get("degStart", degStart);
  Config::Get("degEnd", degEnd);
  Config::Get("Problem",Conf.s_prob );
  vector<FWISeismogram> s_ms;
  auto problem = GetInversionProblem(Conf);
  int nrLvs = meshes->Level() - meshes->pLevel();
  string l;
  std::shared_ptr<FWIObservationSpecification> spec;
  double signalFreq = Conf.SConf.SignalFreq;
  for (int deg = degStart; deg <= degEnd; ++deg) {
    for (int lv = meshes->pLevel(); lv <= meshes->Level(); ++lv) {
      LevelPair lev{lv, -1};
      FWIDiscretizationStack Stack(meshes, problem->getParamComps(), problem->getStateComps(), deg,
                                   lev);
      (*Stack.wD)().PrintInfo();
      (*meshes)[lev].PrintInfo();
      auto eleCache = std::make_unique<EleCacheContainer>(*Stack.wV, problem->getNumL());
      Config::Get("Specfile",l);
      if (l.empty()) {
        spec = FWIObsSpecWPCreator().WithT(Conf.TConf.T).Withdt(Conf.TConf.dt).Witht0(Conf.TConf.t0).CreateShared();
      }else{
        int SamplingFactor = 1;
        Config::Get("SpecfileSamplingFactor",SamplingFactor);
        spec = std::make_shared<FWIObservationSpecification>(Conf.TConf.t0,Conf.TConf.T,Conf.TConf.dt,l.c_str());
      }
      FWIObservationOperatorNEW *Obs = FWIObservationCreator(eleCache->Get(),
                                                             Stack.GetMesh(),
                                                             problem->getNumL()).WithSpec(
          spec).Create();
      std::cout << std::setprecision(15);
      problem->setInitialValueFlag(true);
      auto M = std::make_unique<Vector>(*Stack.mV);
      auto M2 = std::make_unique<Vector>(*Stack.mV);
      M2->Clear();
      M->Clear();
      FillVectorFromProblem(*problem, *M2);
      ReparametrizeMaterialVector(*M2, *problem, *M);
      auto PMLH = std::make_shared<PMLHandler>();
      problem->setShot(0);
      FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
      auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem, *M2);
      auto Mat2 = std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem, *M2);
      auto MatAss = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true,
                                                                 &eleCache->Get());
      auto plotter = std::make_shared<DGVAcousticPlotter>(problem->getNumL(), &eleCache->Get());
      auto problemVA = std::make_shared<FWIProblemWrapperVA>(*problem, *problem, std::move(M),
                                                             problem->getNumL(), problem->getDim(),
                                                             problem);

      auto rhsAssembler = std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
      auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss,
                                                                            rhsAssembler);
      Vectors U_m(Conf.TConf.timeSteps + 1, *Stack.wV);
      FWISeismogram s_m;
      assembleCombined->SetEleCache(&eleCache->Get());
      assembleCombined->SetInitialValue(U_m[0]);
      TSolveR.SolveAndMeasure(assembleCombined, *Obs, s_m, &U_m);
      s_ms.push_back(s_m);
      s_m.collect();
      s_m.writeToFile(l+"S__deg_" + to_string(deg)+"__lv_" + to_string(lv) + "__freq_" + to_string(signalFreq));
    }
  }
  for (int i = 1; i < s_ms.size();++i){
    FWISeismogram diff = s_ms[i-1];
    diff -= s_ms[i];
    MOUT(sqrt(diff * diff));
  }
  return 1;
}
