//#include "m++.hpp"
#include "m++.hpp"
#include <sstream>

#include <iostream>
#include "FWINonlinearSolvers.hpp"
#include "ForwardModels/FWIForwardModelDGTSVA.hpp"
#include "ForwardModels/FWIForwardModelDGTSVAFullWave.hpp"
#include "ForwardModels/FWIForwardModelDGTSVAAllAtOnce.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "SaveLoad.hpp"
#include "AAO/AAOIntegralPrecon.hpp"
#include "FWIHelpers.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "Observation/FWIObservationCreator.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "VTUModelReader.hpp"
#include "FWIVectorFillFunctions.hpp"
#include "SourceInversion.hpp"
#include "ForwardModels/FWIForwardModelDGTSVANormalized.hpp"
#include "FWIProblemWrapperVA.hpp"
//#include "FWIForwardModelDGTS.hpp"

class MeasureConf {
public:
  std::string measurement_time = "point";
  std::string measurement_space = "point";
  std::string ReceiverPlacement = "Waypoints";
};

void FWIInverseMainDGTSVANEWNEW() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  auto prob = GetInversionProblem(Conf);
  LevelPair Lev = {0, -1};
  int deg = 0;
  FWIDiscretizationStack discStack(meshes, prob->getParamComps(), prob->GetStateComponents(), deg,
                                   Lev);
  auto eleCache = std::make_unique<EleCacheContainer>(*discStack.wV, prob->getNumL());

  std::shared_ptr<FWIObservationSpecification> spec = nullptr;
  spec = FWIObsSpecWPCreator().CreateShared();

  auto mask = GetMeasureMaskNEW(meshes->dim(), prob->getNumL(), 0);
//    FWIObservationOperatorNEW obs(std::move(timeOp), std::move(spaceOp), std::move(mask));
    FWIObservationOperatorNEW *Obs = FWIObservationCreator(eleCache->Get(), discStack.GetMesh(), prob->getNumL()).Create();
//    meshes->PrintInfo();
//    (*discStack.mD)().PrintInfo();
//    (*discStack.wD)().PrintInfo();
//
//    std::unique_ptr<FWIForwardModelDGTSVA> FWD(getFWD(meshes, Conf.level, deg, GetInversionProblem(Conf)));
//    auto Psi = configureObs(FWD->GetMesh(), *FWD, Conf);
//    FWD->setObservationOperator(Psi.get());
//    std::string seismopath = "";
//    Config::Get("seismopath", seismopath);
//    if (seismopath == "") {
//        FWIDataContainerDomain real(*FWD);
//        Vector M_real(FWD->GetMatVec());
//        M_real.Clear();
//        FWD->fillVectorFromProblem(M_real, FWD->GetProblem());
//        real.getMatWrite() = M_real;
//        FWD->plot(M_real, "fwi_m_real");
//        TIMEFUNCTION(FWD->createRealData(real, obs);)
//        if (Conf.SaveObservedSeismogram) {
//            FWISeismogram printObs;
//            for (int i = 0; i < obs.size(); ++i) {
//                printObs = obs[i].getSeismogram();
//                printObs.collect();
//                printObs.writeToFile(Config::getDataPath() + string("FWI/Obs") + std::to_string(i));
//            }
//        }
//    } else {
//        int shots = FWD->GetProblem().getShotCount();
//        for (int shot = 0; shot < shots; ++shot) {
//            std::string seismostring = seismopath + std::to_string(shot);
//            mout << "Reading data from " + seismostring + " make sure it has " +
//                    std::to_string(FWD->GetProblem().getParamComps()) + " entries!" << endl;
//
//        }
//    }
}


std::shared_ptr<FWIForwardModelDGTSVA>
getFWDshared(std::shared_ptr<Meshes> M, int level, int deg,
             std::shared_ptr<WaveProblemDGTSVA> prob) {
  std::string type = "";
  Config::Get("ReconstructionType", type);
  LevelPair lev{level, -1};
  if (type == "Classical") {
    return std::make_shared<FWIForwardModelDGTSVA>(M, lev, deg, prob);
  } else if (type == "FullWave")
    return std::make_shared<FWIForwardModelDGTSVAFullWave>(M, lev, deg, prob);
  else if (type == "AllAtOnce")
    return std::make_shared<FWIForwardModelDGTSVAAllAtOnce>(M, lev, deg, prob);
  else if (type == "Normalized")
    return std::make_unique<FWIForwardModelDGTSVANormalized>(M, lev, deg, prob);
  else {
    Exit("specify type or specify a known type");
    return nullptr;
  }
}

std::unique_ptr<FWIForwardModelDGTSVA>
getFWD(std::shared_ptr<Meshes> meshes, int level, int deg,
       std::shared_ptr<WaveProblemDGTSVA> prob) {
  std::string type = "";
  Config::Get("ReconstructionType", type);
  LevelPair lev{level, -1};
  if (type == "Classical") {
    return std::make_unique<FWIForwardModelDGTSVA>(meshes, lev, deg, prob);
  } else if (type == "FullWave")
    return std::make_unique<FWIForwardModelDGTSVAFullWave>(meshes, lev, deg, prob);
  else if (type == "AllAtOnce")
    return std::make_unique<FWIForwardModelDGTSVAAllAtOnce>(meshes, lev, deg, prob);
  else if (type == "Normalized")
    return std::make_unique<FWIForwardModelDGTSVANormalized>(meshes, lev, deg, prob);
  else {
    Exit("specify type or specify a known type");
    return nullptr;
  }
}

void createMeasuredSeismogram(const std::shared_ptr<Meshes> &meshes, int deg, WPConf &Conf,
                              std::vector<FWIDataContainerRange> &obs) {
  bool tmp = Conf.SConf.FrequencyFiltering;
  if (tmp) Conf.SConf.FrequencyFiltering = false;
  auto f_w_d = getFWD(meshes, Conf.level, deg, GetInversionProblem(Conf));
  std::unique_ptr<FWIForwardModelDGTSVA> FWD(std::move(f_w_d));
  std::string seismopath;
  Config::Get("seismopath", seismopath);
  if (seismopath.empty()) {
    FWIDataContainerDomain real(FWD->getReconstructionType(), FWD->GetMatVec(),
                                FWD->GetWaveVec(), FWD->TConf.timeSteps);
    Vector M_real(FWD->GetMatVec());
    M_real.Clear();
    FillVectorFromProblem(FWD->GetProblem(),M_real);
//    FWD->fillVectorFromProblem(M_real, FWD->GetProblem());
    real.getMatWrite() = M_real;
    FWD->plot(M_real, "fwi_m_real");
    double signalToNoise = -100000;
    Config::Get("SignalToNoise",signalToNoise);
    int  signalToNoiseSeed = 42;
    Config::Get("SignalToNoiseSeed",signalToNoiseSeed);
    TIMEFUNCTION(FWD->createRealData(real, obs);)
    if (Conf.SConf.SaveObservedSeismogram) {
      FWISeismogram printObs;
      for (int i = 0; i < obs.size(); ++i) {
        printObs = obs[i].getSeismogram();
        printObs.collect();
        printObs.writeToFile(Config::GetDataPath() + string("FWI/Obs") + std::to_string(i));
      }
    }
    if (signalToNoise > -100000+Eps) {
      int seisnr = 0;
      for (auto &seis: obs) {
        seis.getSeismogramWrite().addRelativeNoise(signalToNoise, signalToNoiseSeed+ seisnr++*1000000);
      }
      if (Conf.SConf.SaveObservedSeismogram) {
        FWISeismogram printObs;
        for (int i = 0; i < obs.size(); ++i) {
          printObs = obs[i].getSeismogram();
          printObs.collect();
          printObs.writeToFile(Config::GetDataPath() + string("FWI/ObsNoisy") + std::to_string(i));
        }
      }
    }
  } else {
    string format = "FWI";
    Config::Get("ReadFormat", format);
      int shots = FWD->GetProblem().getShotCount();
    if (format == "FWI") {
      const auto & spaceOp = FWD->GetObservationOperator().GetSpaceOp();
      for (int shot = 0; shot < shots; ++shot) {
        std::string seismostring = seismopath + std::to_string(shot);
        mout << "Reading data from " + seismostring + " make sure it has " +
                std::to_string(FWD->GetProblem().getParamComps()) + " entries!" << endl;
        obs.emplace_back(FWISeismogram(seismostring.c_str(), 0,spaceOp.GetSpaceSupport()));
      }
    }else if (format == "IFOS"){
      string specf, seismoname;
//      Config::Get("Specfile",specf);
      Config::Get("seismoname",seismoname);
//      FWIObservationSpecification spec(specf.c_str());
      const auto & spaceOp = FWD->GetObservationOperator().GetSpaceOp();
      char data[70];
      for (int shot = 1; shot <= shots; ++shot) {
        sprintf(data, "%s%d", (seismopath + seismoname).c_str(), shot);
        FWISeismogram tmpSeis(FWD->GetObservationOperator().GetSpaceOp().GetSpecification(),
                              spaceOp.GetSpaceSupport(),data);

        obs.emplace_back(tmpSeis);

      }
    }
  }
  if (tmp) Conf.SConf.FrequencyFiltering = true;
}

void fillInitialModel(FWIForwardModelDGTSVA &FWD, FWIDataContainerDomain &initial) {
  bool loadInitialModelFromFile = false;
  bool loadRealModelFromFile = false;
  Config::Get("loadInitialModelFromFile", loadInitialModelFromFile);
  Config::Get("loadRealModelFromFile", loadRealModelFromFile);
  Vector M(FWD.GetMatVec());
  M.Clear();
  FWD.GetProblemWrite().setInitialValueFlag(true);
  FillVectorFromProblem(FWD.GetProblem(), M);

//  FWD.fillVectorFromProblem(M, FWD.GetProblem());
  if (FWD.GetProblem().getDim() == 2) {
    FWD.GetProblemWrite().setInitialValueFlag(false);
    FWD.fillSourceAndReceiverCellsFromProblem(M, FWD.GetProblem());
    FWD.GetProblemWrite().setInitialValueFlag(true);
  }
  if (FWD.PMLHAAO.get()) {
    FWD.PMLHAAO->SetPMLZero(M);
  }
  if (loadInitialModelFromFile) {
    std::vector<std::string> fnv;
    std::string fn = "";
    Config::Get("InitialModelFileName", fn);
    if (fn == "") Exit("Set Initial Model Filename!")
    fnv.push_back(fn);
    Vector Tmp(M);
    Tmp.Clear();
    VTUModelReader reader(fnv, M, "Vp");
    double VpMin = 1000000.0;
    double VpMax = -1000000.0;
    Config::Get("InitialModelBoxLower", VpMin);
    Config::Get("InitialModelBoxUpper", VpMax);
    if (fn.find("Exp") == std::string::npos) {
      std::string s = "Vp";
      reader.getData(Tmp, s, 1);
    } else {
      std::string s = "VpExp";
      reader.getData(Tmp, s, 1);
    }
    if (FWD.GetProblem().getParametrization() == "taup") {
      for (cell c = M.cells(); c != M.cells_end(); ++c) {
        double tmp = Tmp(c(), 1);
        tmp = max(tmp, VpMax);
        tmp = min(tmp, VpMin);
        M(c(), 1) = tmp;
      }
    } else if (FWD.GetProblem().getParametrization() == "taup_positive") {
      for (cell c = M.cells(); c != M.cells_end(); ++c) {
        double tmp = Tmp(c(), 1);
        tmp = max(tmp, VpMax);
        tmp = min(tmp, VpMin);
        M(c(), 1) = log(tmp);
      }
    }
  }
    FWD.plot(M, "fwi_m_initial");
    initial.getMatWrite() = M;
}

void
processInitialValue(FWIForwardModelDGTSVA &FWD, std::shared_ptr<FWIDataContainerDomain> &initial,
                    WPConf Conf) {
  auto initialnew = std::make_shared<FWIDataContainerDomain>(FWD.getReconstructionType(),
                                                             FWD.GetMatVec(),
                                                             FWD.GetWaveVec(), FWD.TConf.timeSteps);
  FWD.projectModel(initial->getMat(), initialnew->getMatWrite());
  initial = initialnew;
  FWD.smoothMaterial(initial->getMatWrite(), Conf.SConf.MaterialSmoothPasses);
}

vector<NonlinearOutput> FWIInverseMainDGTSVA() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  LevelPair lev{Conf.level, -1};
  int deg = 0;
  Config::Get("deg", deg);
  std::vector<NonlinearOutput> Output;
  bool firstPass = true;
  std::shared_ptr<FWIDataContainerDomain> initial = nullptr;
  std::vector<FWIDataContainerRange> obs;
  createMeasuredSeismogram(meshes, deg, Conf, obs);
  std::shared_ptr<FWIForwardModelDGTSVA> FWD;
  std::vector<FWISeismogram> ObsSeismos;
  auto prob = GetInversionProblem(Conf);
  if (Conf.SConf.FrequencyFiltering) {
    for (int i = 0; i < obs.size(); ++i) {
      ObsSeismos.push_back(obs[i].getSeismogram());
    }
  }
  for (int j = 0; j < Conf.SConf.FrequencyLevels.size(); ++j) {
    Conf.SConf.CurrentLevel = j;
    auto freq = Conf.SConf.getFreq();
    mout << "Frequency Filter to: " << freq << endl;
    if (Conf.SConf.FrequencyFiltering) {
      for (int i = 0; i < obs.size(); ++i) {
        obs[i].getSeismogramWrite() = ObsSeismos[i];
        if (Conf.SConf.FilteringType == "LowPass") obs[i].getSeismogramWrite().LowPassFilter(freq[0], Conf.SConf.FilteringOrder);
        else if (Conf.SConf.FilteringType == "BandPass"){
          obs[i].getSeismogramWrite().BandPassFilterIFOS(freq[0],freq[1]);
        }
        if (Conf.SConf.SaveFilteredSeismogram) {
          FWISeismogram obsSeis = obs[i].getSeismogram();
          obsSeis.collect();
          string s;
          for (double d : freq){
            s += std::to_string(d);
          }
          obsSeis.writeToFile(Config::GetDataPath() + string("FWI/ObsFilteredShot") + to_string(i)  +
                                  Conf.SConf.FilteringType + s);
        }
      }
      if(!Conf.SConf.SrcInversion){
        prob->SetTimeSource(GetSourceWavelet(Conf.TConf,Conf.SConf));
      }
    }
    auto FWDNew = getFWDshared(meshes, Conf.level, deg, prob);
    if (!firstPass) {
      processInitialValue(*FWDNew, initial, Conf);
    }
    FWD = FWDNew;
    if (Conf.SConf.FrequencyFiltering)  FWD->SetPreconRegularization(Conf.SConf.PreconRegu[j]);

    std::cout << std::setprecision(15);
    if (firstPass) {
      initial = std::make_shared<FWIDataContainerDomain>(FWD->getReconstructionType(),
                                                         FWD->GetMatVec(),
                                                         FWD->GetWaveVec(), FWD->TConf.timeSteps);
      fillInitialModel(*FWD, *initial);
      firstPass = false;
    }
    if(Conf.SConf.SrcInversion)
    {
      std::vector<FWIDataContainerRange> obsTmp;
      FWD->createRealData(*initial, obsTmp);
      SourceInversionProblem(obs,obsTmp,FWD->problem_fwd,FWD->GetObservationOperator(),
                             Conf.SConf.SrcInversionEps, Conf.SConf.SrcInversionKappa,j);
    }
    std::unique_ptr<FWINonlinearSolver> NLS(GetFWINonlinearSolver(*FWD));
    std::string suffix;
    if (Conf.SConf.FrequencyFiltering) {
      for (double d: freq) {
        suffix += std::to_string((int) (d * 100));
      }
    }
    Output.push_back((*NLS)(obs, *initial, suffix));
    if(Conf.SConf.FrequencyFiltering){
      mout << "Loop over" + suffix + "ended" << endl;
      mout << "The residuals are:" << endl;
      const NonlinearOutput& CurrentOutput = Output.back();
      for (int shot = 0; shot< CurrentOutput.size(); ++shot){
        mout << "Residuals of shot " + to_string(shot) << endl;
          for (const auto& Result : CurrentOutput[shot]){
              mout << Result << endl;
          }
      }

    }
  }
  return Output;
}

int createKernels2() {
  auto meshes = getFWIMesh();
  WPConf Conf;
  LevelPair lev{Conf.level, -1};
  int deg = 0;
  int numL=0;
  Config::Get("numL",numL);

  Config::Get("deg", deg);
  Conf.s_prob = "VAcousticWaveC2Benchmark";
  std::shared_ptr<WaveProblemDGTSVA> problem = GetInversionProblem(Conf);
  FWIDiscretizationStack Stack(meshes, problem->getParamComps(), problem->getStateComps(), deg,
                               lev);
  (*Stack.wD)().PrintInfo();
  auto eleCache = std::make_unique<EleCacheContainer>(*Stack.wV, problem->getNumL());
  auto spec = FWIObsSpecWPCreator().WithT(Conf.TConf.T).
      Withdt(Conf.TConf.dt).Witht0(Conf.TConf.t0).CreateShared();
  FWIObservationOperatorNEW *Obs = FWIObservationCreator(eleCache->Get(),
                                                         Stack.GetMesh(),
                                                         problem->getNumL()).WithSpec(
      spec).Create();
  std::cout << std::setprecision(15);
  problem->setInitialValueFlag(true);
  auto M = std::make_unique<Vector>(*Stack.mV);
  auto M2 = std::make_unique<Vector>(*Stack.mV);
  M2->Clear();
  M->Clear();
  FillVectorFromProblem(*problem, *M2);
  ReparametrizeMaterialVector(*M2, *problem, *M);
  auto PMLH = std::make_shared<PMLHandler>();

    problem->setShot(0);
    FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
    auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem, *M2);
    auto Mat2 = std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem, *M2);
    auto MatAss = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true, &eleCache->Get());
    auto plotter = std::make_shared<DGVAcousticPlotter>(problem->getNumL(), &eleCache->Get());
    std::shared_ptr<FWIProblemWrapperVA> problemVA = std::make_shared<FWIProblemWrapperVA>(*problem, *problem,
                                                                                           std::move(M),
                                                                                           problem->getNumL(),
                                                                                           problem->getDim(),
                                                                                           problem);

    auto rhsAssembler = std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
    auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss, rhsAssembler);
    Vectors U_m(Conf.TConf.timeSteps + 1, *Stack.wV);
    U_m.Clear();
    assembleCombined->SetEleCache(&eleCache->Get());
    assembleCombined->SetInitialValue(U_m[0]);
    TSolveR.Solve(assembleCombined, U_m);
    reconstructionContainer recon;
    recon.Vp = true;
    recon.Rho = true;
    recon.Taup = true;
    FWILinOpAdjRed AdjOp(U_m, *M2, *Obs, problem, PMLH, Conf.TConf, recon, eleCache->Get());
    FWISeismogram Seis(*spec, 1);
    Vector Eval(*Stack.mV);
    Eval.Clear();
    int nSteps = double(spec->nSteps());
    int KernelStep = -1.0;
    int KernelSkip = 10;
    Config::Get("KernelEveryStep", KernelStep);
    Config::Get("KernelSkip", KernelSkip);
  if (KernelStep < 0) Exit("specify \"KernelEveryStep\" option");
    int countMax = int(floor(nSteps / double(KernelStep)));
    for (int count = KernelSkip; count < countMax; ++count) {
        Seis.ts_clear();
        for (auto receiver = spec->begin(); receiver != spec->end(); ++receiver) {
            for (int i = 0; i < spec->nSteps(); ++i) {
                double t = spec->step(i);
                FWIState Measure(0.0, 1);
                if (i == (KernelStep * count)) {
                    Measure[0] = 1.0;
                }
                Seis.addMeasurement(*receiver, t, Measure);
            }
        }
        AdjOp.CalculateAdjoint(Seis, Eval, nullptr);
      auto &Plot = mpp::plot("Kernel_" + to_string(count));
      Plot.AddData("Rho", Eval, 0);
      Plot.AddData("Vp", Eval, 1);
      if (numL>0){
        Plot.AddData("Taup", Eval, 2);
      }
      Plot.PlotFile();

  }
  return 0;
}

