#ifndef FWIFORWARDMODEL
#define FWIFORWARDMODEL

#include "lib1_math/basicalgebra/RMatrix.hpp"
#include "lib4_fem/elements/Element.hpp"
#include <memory>
#include "m++.hpp"

#include "FWIHelpers.hpp"
#include "Observation/FWISeismogram.hpp"
#include "Assemble.hpp"
#include "AlgebraFwd.hpp"
#include "PMLHandler.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"

typedef RVector FWIState;

typedef std::vector<FWIDataContainerRange> FWISubproblem;

class FWIMaterial {
public:
    virtual ~FWIMaterial() = default;

};

class FWIBndConditions {
public:
    virtual ~FWIBndConditions() = default;

};

class FWIRhs {
protected:
    int shot_count;
public:
    explicit FWIRhs(int shot_count = 1, int shot = 0) : shot_count(shot_count) {}

    virtual ~FWIRhs() = default;

    virtual int getShotCount() const { return shot_count; }


    virtual void setShot(int _shot) {Exit("CIAO"); }

    virtual int getShot() const {Exit("CIAO"); }
};


class FWISeismogram;

// TODO: Move operator() to ForwardModel
class FWIForwardModel;


class FWIForwardModel {
protected:
    FWIDiscretizationStack FWIDiscStack;
    std::shared_ptr<PMLHandler> PMLH = nullptr;
  bool useGardener = false;

public:
  std::shared_ptr<PMLHandlerAAO> PMLHAAO = nullptr;
    TimeConf TConf;
    std::unique_ptr<Matrix> MPC = nullptr;

    FWIForwardModel() = default;

    const Vector &GetWaveVec() const {
        return *FWIDiscStack.wV;
    }

    const Vector &GetMatVec() const {
        return *FWIDiscStack.mV;
    }

    virtual ~FWIForwardModel() {}

    virtual const char *Name() const = 0;

    virtual bool GetPrecon() const {
        Exit("Set Preconstatus")
    }
  virtual bool UseBoxConstraint() {return false;}
  virtual void ApplyBoxConstraint(FWIDataContainerDomain&,FWIDataContainerDomain&) {Exit(("CIAO"));}

  virtual void Gardener(FWIDataContainerDomain&,FWIDataContainerDomain&){ Exit("implement this in your fwdmodel")}
  virtual bool UseGardener(){ return useGardener;}

  virtual std::string getReconstructionType() const { return "Classical"; }

    virtual auto getReconstructionlist() -> std::list<unsigned int> const {
        Exit("not implemented, location of parameters to be reconstructed within")
    }

    virtual int StateComponents() const = 0;

    virtual int ParameterComponents() const {
        Exit("Not implemented")
        return 0;
    }


    virtual void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                                     FWISubproblem &evaluation,
                                     std::vector<int> subproblem) = 0;

    virtual void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                                   const FWISubproblem &, FWIDataContainerDomain &) = 0;

    virtual void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                            const FWIDataContainerDomain &, FWISubproblem &) = 0;

    virtual void calculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                                 const FWISubproblem &nonlinearEvaluation,
                                 FWISubproblem &defect,
                                 std::vector<int> subproblem) = 0;

    virtual void mSDProcedure(const FWIDataContainerDomain &, FWIDataContainerDomain &, FWIDataContainerDomain &,
                              FWISubproblem &) {Exit("notimplemented")}

    virtual double scalarProductL2(const FWISubproblem &u,
                                   const FWISubproblem &v) const = 0;

    virtual void Parameter2SeismogramMap(const Vector &, FWISeismogram &, int, Vectors *) = 0;

  virtual double scalarProductL2(const FWIDataContainerRange &u,
                                   const FWIDataContainerRange &v) const {
        Exit("needs to be implemented")
        return -1.0;
    }

    virtual double scalarProductL2(const FWIDataContainerDomain &,
                                   const FWIDataContainerDomain &) const {
        Exit("needs to be implemented")
        return -1.0;
    }

    virtual void plot(const Vectors &u,
                      const std::string &name,
                      int nr = -1,
                      int shot = -1) {
        plot(u, name.c_str(), nr, shot);
    }

    virtual void plot(const Vector &m,
                      const std::string &name,
                      int nr = -1,
                      int lvl = -1) { plot(m, name.c_str(), nr, lvl); }

    virtual void plot(const FWIDataContainerDomain &m,
                      const std::string &name,
                      int nr = -1,
                      int lvl = -1) { plot(m, name.c_str(), nr, lvl); }

    virtual void plot(const Vector &m,
                      const char *,
                      int nr = -1,
                      int lvl = -1) {Exit("not implemented. change of interface ")}


    virtual void plot(const FWIDataContainerDomain &,
                      const char *,
                      int nr = -1,
                      int lvl = -1) {Exit("not implemented. change of interface ")}

    virtual void plot(const Vectors &u,
                      const char *name,
                      int nr = -1,
                      int shot = -1) {Exit("not implemented. change of interface ")}

    virtual void initializePrecon(const Vector &M, const Vector &M_update,
                                  const Vectors &U_M);

    virtual void initializePrecon(const FWIDataContainerDomain &m,
                                  const FWIDataContainerDomain &m_update,
                                  const FWISubproblem &u_m);


    virtual void applyPrecon(const FWIDataContainerDomain &m,
                             FWIDataContainerDomain &);

  virtual void SetPenalty(const Vector &vector, FWISubproblem &vector1);

    virtual void GetInitialValueForInner(const FWIDataContainerDomain &iterateNonlinear,
                                         const FWISubproblem &nonlinearEvaluation,
                                         FWISubproblem &defectLinearized,
                                         FWIDataContainerDomain &iterateMUpdate,
                                         std::vector<int> prob) {}

    virtual int numberOfShots() const = 0;


};

FWISubproblem &operator*=(FWISubproblem &vec, double d);

FWISubproblem &operator-=(FWISubproblem &vec1, FWISubproblem &vec2);

FWISubproblem &operator+=(FWISubproblem &vec1, FWISubproblem &vec2);

#endif
