#ifndef _FWI_ABSTRACTINTERFACEDGTS_
#define _FWI_ABSTRACTINTERFACEDGTS_

#include "FWIDataContainerRange.hpp"
#include "AlgebraFwd.hpp"
#include "Assemble.hpp"
#include "Observation/FWISeismogram.hpp"
#include "FWIHelpers.hpp"
#include "m++.hpp"
#include "lib4_fem/elements/Element.hpp"
#include "lib1_math/basicalgebra/RMatrix.hpp"
#include <iomanip>
#include <utility>
#include "BasicDoFs.hpp"
#include "FWIForwardModel.hpp"
#include "SaveLoad.hpp"
#include "FWIMppHelpers.hpp"
#include "MeshesCreator.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "normCalculator.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "FWIAdjointCalculator.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "AAO/FWIIntegralOperator.hpp"
#include "GramianBlockInverse.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "TimeIntegrator.hpp"
#include "FWIDGVAcousticMaterialFromVector.hpp"
#include "Assembler/FWIRhsAssemblerLinearized.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "VectorsIntegrator.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "BasicSolver.hpp"
#include "DGVAcousticAssembleBwd.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "FWIDataContainerDomain.hpp"
#include "Observation/FWIObservationCreator.hpp"
#include "FWITaper.hpp"
#include "Reduced/FWILinOpAdjRed.hpp"
#include "Reduced/FWILinOpRed.hpp"


class FWIForwardModelDGTS : public FWIForwardModel {
protected:
    std::vector<std::unique_ptr<Vectors>> uM;
    std::unique_ptr<FWIObservationOperatorNEW> Obs = nullptr;
    reconstructionContainer recon;
    std::unique_ptr<FWILinOpAdjRed> AdjOp;
    std::unique_ptr<FWILinOpRed> LinOp;
public:
    mutable int cntnr = 0;
    std::shared_ptr<WaveProblemDGTSVA> problem_fwd = nullptr;
    int paramComps = -1;
    int stateComps = -1;
    int numL = -1;
    int deg = -1;
    std::shared_ptr<DGVAcousticPlotter> VAPlotter = nullptr;
    std::shared_ptr<DGVAcousticRHSAssemblerAdjoint> RHSAdjoint = nullptr;
    std::unique_ptr<Vector> MPCVec = nullptr;
    std::unique_ptr<EleCacheContainer> eleCacheContainer;
    std::unique_ptr<MeasureMask> Mask = nullptr;
    bool allowPlotting_M = true;
    bool allowPlotting_WF = false;
    int plotFactor = 1;
    int plotEnding = -1;
    bool taper = false;
    bool dampingFlux = false;
    std::list<unsigned int> reconlist;
    std::vector<unsigned int> reconVec;
    double taperExponent = -5;
    std::pair<bool, int> precon = {false, 0};
    std::unique_ptr<FWITaper> Taper = nullptr;
    double PreconRegu{1e-08};
    double CellWiseTimeProduct(const Vectors &U, const Vectors &V, const DGAcousticElement &elem, double dt) const;

    double CellWiseStepProduct(const Vector &U, const Vector &V, const DGAcousticElement &elem,
                               double dt_) const;

    void initializeMeasureMask();

    void initializeRecon();

// --------------------------------------------------------------
// Debugging and testing

  void fillSourceAndReceiverCellsFromProblem(Vector &M, const WaveProblemDGTSVA &prob) const;

public:
    FWIForwardModelDGTS(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                        const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd)
            : FWIForwardModel(),
               deg(deg) {
        FWIDiscStack.meshes = std::move(meshes);
        FWIDiscStack.lev = lev;
        Config::Get("FWIPrecon", precon.first);
        Config::Get("FWIPreconType", precon.second);
        Config::Get("allowPlotting_M", allowPlotting_M);
        Config::Get("allowPlotting_WF", allowPlotting_WF);
        Config::Get("plotFactor", plotFactor);
        Config::Get("dampingFlux", dampingFlux, -1);

        if (!(problem_fwd == nullptr)) {
            setProblem(problem_fwd);
        }

    }
    bool UseBoxConstraint() override {
      if(Taper) {
        return Taper->Conf.UseMinMaxTaper;
      }else
        return false;
    }
  void ApplyBoxConstraint(FWIDataContainerDomain& iterateNonlinear,
                          FWIDataContainerDomain& iterateMUpdate) override {
      Taper->ApplyMinMaxTaperAddition(iterateNonlinear.getMatWrite(),iterateMUpdate.getMatWrite(),*problem_fwd);
  }
    const WaveProblemDGTSVA &GetProblem() { return *problem_fwd; }

    WaveProblemDGTSVA &GetProblemWrite() { return *problem_fwd; }

    // --------------------------------------------------------------
    // initialization methods
    void initializeSpaceDiscretization(FWIDiscretizationStack &discStack);

    virtual void setProblem(std::shared_ptr<WaveProblemDGTSVA> _prob);

    bool UseGardener(){return useGardener;}
    int GetDim() const;

    int numberOfShots() const override { return problem_fwd->getShotCount(); }

    auto getReconstructionlist() -> std::list<unsigned int> const override { return reconlist; }

  const FWIObservationOperatorNEW& GetObservationOperator(){
      return *Obs;
    }
void SetPreconRegularization(double reg){
    PreconRegu =reg;
    }
  void makeFD(const Vectors &U_m, Vectors &U_m_fd) const;

    int StateComponents() const override { return stateComps; }

    int ParameterComponents() const override { return paramComps; }


  double L2TScalar(const Vectors &u, const Vectors &v) const;


    void applyPrecon(const Vector &m, Vector &m_update) {
        Vector tmp(m_update);
        m_update = (*MPC) * tmp;
    }

    void applyPrecon(const FWIDataContainerDomain &m,
                     FWIDataContainerDomain &m_update) override {
      applyPrecon(m.getMat(), m_update.getMatWrite());
    }

    void assemblePseudoHessianPrecon(const Vector &m, const Vector &m_update,
                                     const Vectors &u, const Vectors &u_fd,
                                     Matrix &PHess,
                                     Vector *Plotti = nullptr) const;


    double L2tScalar(const Vector &u, const Vector &v) const;


  double NormL2(const Vectors &U) const { return sqrt(L2TScalar(U, U)); }

  double NormL2(const Vector &U) const { return sqrt(L2Scalar(U, U)); }

    double L2Scalar(const Vector &u,
                    const Vector &v) const;

    virtual std::shared_ptr<DGAssemble> getDGAssemble(const Vector &m,
                                                      const FWIBndConditions &bnd,
                                                      const FWIRhs &rhs,
                                                      bool fwd) = 0;


    void projectModel(const Vector &U2, Vector &UM) const;

    /* Applying this method with passes=2 is equivalent to convolving with the following kernel
     *      1
     *    2 4 2
     *  1 4 8 4 1
     *    2 4 2
     *      1
     */
    void smoothMaterial(Vector &m, int passes = 1) const {
        for (int i = 0; i < passes; ++i) {
            __smoothMaterial(m);
        }
    }

    /* smoothing with kernel
     *    1
     *  1 2 1
     *    1
     */
    void __smoothMaterial(Vector &UM) const {
        Vector UMM(UM);
        UMM.Accumulate();

        const Mesh &M = UM.GetMesh();
//    int paramComps = 2;
        UM.Clear();
        for (cell c = UM.cells(); c != UM.cells_end(); ++c) {
            double counter = 0;
            double fac = 0;

            // cell itself
            fac = 2;
            for (int i = 0; i < paramComps; ++i) {
                UM(c(), i) += fac * UMM(c(), i);
            }
            counter += fac;

            // face-neighbors
            for (int f = 0; f < c.Faces(); ++f) {
                Point N = Neighbor(M, *c, f);
                if (N != Infty) {
                    fac = 1;
                    for (int i = 0; i < paramComps; ++i) {
                        UM(c(), i) += fac * UMM(N, i);
                    }
                    counter += fac;
                }
            }
            // normalize
            for (int i = 0; i < paramComps; ++i) {
                UM(c(), i) /= counter;
            }
        }
        UM.MakeAdditive();
  }

    virtual bool GetPrecon() const override {
        return precon.first;
    }

    virtual void calculateL2AdjointStep(const Vector &,
                                        const Vectors &,
                                        int,
                                        const Vector &,
                                        Vector &) const {
        Exit("not implemented for this model");
    }


    virtual ~FWIForwardModelDGTS() {
    }

    void Parameter2SeismogramMap(const Vector &m, FWISeismogram &s, int shot, Vectors *U_m) final;

};

#endif
