#ifndef _FWIFORWARDMODELDGTSVA_H_
#define _FWIFORWARDMODELDGTSVA_H_

#include "FWIForwardModel.hpp"
#include "FWIAdjointCalculator.hpp"

#include "Problems/FWIProblemsDGTSVA.hpp"
#include <iomanip>
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "AAO/FWIIntegralOperator.hpp"
#include "GramianBlockInverse.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"


class FWIForwardModelDGTSVA : public FWIForwardModelDGTS {
public:
  virtual void MakeLinearAndAdjointOperator(const Vector& m, Vectors& uM){
    mout << "Assembling Linearized Operators" << endl;
    AdjOp = std::make_unique<FWILinOpAdjRed>(uM, m, *Obs, problem_fwd, PMLH, TConf, recon,
                                             eleCacheContainer->Get());
    LinOp = std::make_unique<FWILinOpRed>(uM, m, problem_fwd, PMLH, TConf, eleCacheContainer->Get(),
                                          false);
    mout << "Done!" << endl;
  }
  virtual void Nonlinear(const Vector& m, Vectors& uM,FWISeismogram &sM,int shot){
    Parameter2SeismogramMap(m,sM,shot,&uM);
//    Parameter2SeismogramMap(nonlinearPoint.getMat(), evaluation[i].getSeismogramWrite(), subproblem[i],
//                            uM[i].get());
  }
    FWIForwardModelDGTSVA(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                          const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd);

    virtual void createRealData(const FWIDataContainerDomain &m,
                                std::vector<FWIDataContainerRange> &obs) {
        Vectors tmp(TConf.timeSteps + 1, GetWaveVec());
        tmp.Clear();
        for (int shot = 0; shot < (*this).problem_fwd->getShotCount(); ++shot) {
            mout << "Generating Real data for shot " << shot << endl;
            obs.emplace_back(getReconstructionType(), GetWaveVec(), TConf.timeSteps);
//            obs[shot].resetWaveData();
            Parameter2SeismogramMap(m.getMat(), obs[shot].getSeismogramWrite(), shot, nullptr);
//            obs[shot].getSeismogramWrite().writeToFile("data/FWI/generatedData_" + std::to_string(shot));
        }
    }

    const Mesh &GetMesh() {
        if (PMLH->isPML()) return PMLH->GetDiscStack().meshes->fine();
        return (*FWIDiscStack.meshes)[FWIDiscStack.lev];
    }

  const char *Name() const override { return "FWDModelVAcoustic"; }

    std::shared_ptr<ProblemVAcoustic> GetWrappedProblem(const Vector &m,
                                                        const FWIBndConditions &bnd,
                                                        const FWIRhs &rhs,
                                                        bool fwd);

    std::shared_ptr<DGAssemble> getDGAssemble(const Vector &m,
                                              const FWIBndConditions &bnd,
                                              const FWIRhs &rhs,
                                              bool fwd) override;

    void plotMatVector(const Vector &, const std::string &);

    // Helper

    void plot(const Vectors &, const char *, int = -1, int  = -1) override;

    void plot(const Vector &, const char *, int = -1, int = -1) override;

    void plot(const FWIDataContainerDomain &, const char *, int= -1, int = -1) override;

   void Gardener(FWIDataContainerDomain& NonLinear,FWIDataContainerDomain& Update) override{
    const Vector& UpdVec = Update.getMat();
    Vector& NonLinearVec = NonLinear.getMatWrite();
       for (cell c = UpdVec.cells(); c != UpdVec.cells_end();++c) {
         double VpUpd = UpdVec(c(),1);
         if(abs(VpUpd) < Eps) continue;
         NonLinearVec(c(),1) += UpdVec(c(),1);
         double Vp = NonLinearVec(c(),1);
         if(problem_fwd->getParametrization() == "taup"){
           NonLinearVec(c(),0) = 310.0 * pow(Vp,0.25);
         }else if(problem_fwd->getParametrization() == "taup_positive"){
           NonLinearVec(c(),0) = log(310) + 0.25*Vp; //  log(310.0 * pow(exp(Vp),0.25))
         }else {
           Exit("CIAO");
         }
       }
     }


    void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                             FWISubproblem &evaluation,
                             std::vector<int> subproblem) override;

    void initializePrecon(const Vector &m, const Vector &m_update,
                          const Vectors &u_m) override;

    void initializePrecon(const FWIDataContainerDomain &m,
                          const FWIDataContainerDomain &m_update,
                          const FWISubproblem &u_m) override {
        if (u_m.size() > 1) Exit("only for kaczmarz implemented");
        initializePrecon(m.getMat(), m_update.getMat(), *uM[0]);
    }

    void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                           const FWISubproblem &adjointLinearizedPoint,
                           FWIDataContainerDomain &adjointLinearizedEvaluation) override;

    void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                    const FWIDataContainerDomain &linearizedPoint,
                    FWISubproblem &linearizedEvaluation) override;

    void calculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                         const FWISubproblem &nonlinearEvaluation,
                         FWISubproblem &defect,
                         std::vector<int> subproblem) override;


    virtual ~FWIForwardModelDGTSVA();

    double scalarProductL2(const FWIDataContainerRange &u,
                           const FWIDataContainerRange &v) const override {
        const auto &s1 = u.getSeismogram();
        const auto &s2 = v.getSeismogram();
        return s1 * s2;
    }

    double scalarProductL2(const FWISubproblem &u,
                           const FWISubproblem &v) const override {
        double skp = 0;
        for (size_t i = 0; i < u.size(); ++i) {
            skp += scalarProductL2(u[i], v[i]);
        }
        return skp;
    }

    double scalarProductL2(const FWIDataContainerDomain &u,
                           const FWIDataContainerDomain &v) const override {
        const auto &UU(u.getMat());
        const FWIDataContainerDomain &vv(v);
//        vv *= SKPDomain;
        const auto &VV(vv.getMat());
        return L2Scalar(UU, VV);
    }


};



#endif

