#ifndef _FWIFORWARDMODELDGTSVAALLATONCE_H_
#define _FWIFORWARDMODELDGTSVAALLATONCE_H_

#include "FWIForwardModelDGTSVA.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "AAO/FWIIntegralOperator.hpp"
#include "GramianBlockInverse.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"

RVector fillShapeValues(const DGAcousticElement &elem, const Point &receiverPoint);

class FWIForwardModelDGTSVAAllAtOnce : public FWIForwardModelDGTSVA {
  std::shared_ptr<Matrix> scalarProductMatrix = nullptr;
  std::shared_ptr<GramianBlockInverse> GInv = nullptr;
  double lambdaPenalty = 1.0;
  bool discrete = false;
  bool adaptivePenalty = false;
  double safetyFactor = 1.0;
  bool AAOVerbose = false;
protected:
public:
  FWIForwardModelDGTSVAAllAtOnce(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                                 std::shared_ptr<WaveProblemDGTSVA> problem_fwd_,
                                 bool _forwardOnly = false) : FWIForwardModelDGTSVA(meshes, lev,
                                                                                    deg,
                                                                                    problem_fwd_) {
    Config::Get("LambdaPenalty", lambdaPenalty);

    Config::Get("adaptivePenalty", adaptivePenalty);
    Config::Get("safetyFactor", safetyFactor);
    Config::Get("AAOVerbose", AAOVerbose);

    Config::Get("discrete", discrete);
    Config::Get("Precon", precon.first);
    Config::Get("PreconVers", precon.second);

    Vector vectorTmp(GetWaveVec());
    if (scalarProductMatrix == nullptr) {
      scalarProductMatrix.reset(new Matrix(vectorTmp));
      assembleScalarProductMatrix(*scalarProductMatrix, eleCacheContainer->Get());
      GInv = std::make_shared<GramianBlockInverse>(scalarProductMatrix, eleCacheContainer->Get());
    }
    bool useAAOPML = false;
    Config::Get("UseAAOPML", useAAOPML);
    if(useAAOPML) {
      AAOPMLConfig Conf;
      Vector VEC(0.0,*FWIDiscStack.wD,FWIDiscStack.lev);
      int ts=0;
      Config::Get("PMLTaperStyle", ts);
      double alpha1 = numL>0 ? problem_fwd->alpha1() : -1;
      double alpha2 = numL>0 ? problem_fwd->alpha2() : -1;
      Conf.SetTaperStyle(ts,alpha1,alpha2);
      PMLHAAO = std::make_shared<PMLHandlerAAO>(Conf,VEC,problem_fwd_->getNumL());
    }
  }

  std::shared_ptr<GramianBlockInverse> GetGramianBlockInverse() {
    return GInv;
  }


  std::string getReconstructionType() const override { return "AllAtOnce"; }

  void applyInverseGramian(Vectors &WaveFieldEvaluation) const;

  void calculateDefectWaveField(const Vectors &nonlinearEvaluationWF, Vectors &defect);

  void calculateRHS(Vectors &RhsWF, bool fwd);

  void calculateRHSGInverse(Vectors &rhsWF, bool fwd);

  void calculateRHSRK(Vectors &RhsWF, bool fwd);

  void calculateRHSRKGInverse(Vectors &RhsWF, bool fwd);

  void AdjointMeasurement(const FWISeismogram &Seis, Vectors &adjointLinearizedEvaluationWF);

  void AdjointMeasurement(const FWIDataContainerRange &adjointLinearizedPoint,
                          FWIDataContainerDomain &adjointLinearizedEvaluation);

  void calculateOneDefect(const FWIDataContainerRange &data_obs,
                          const FWIDataContainerRange &nonlinearEvaluation,
                          FWIDataContainerRange &defect);

  void calculateDiscreteLinearizedContribution(const Vector &mDirec, const Vector &Material,
                                               const Vectors &integratedVecs, Vectors &Eva);

  void
  createRealData(const FWIDataContainerDomain &m, std::vector<FWIDataContainerRange> &obs) override;

  const char *Name() const override { return "FWIForwardModelDGTSVAAllAtOnce"; }


  void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                           FWISubproblem &evaluation,
                           std::vector<int> subproblem) override;


  void calculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                       const FWISubproblem &nonlinearEvaluation,
                       FWISubproblem &defect,
                       std::vector<int> subproblem) override;

  void evaluateIntegralOperator(const FWIDataContainerDomain &nonlinearPoint,
                                Vectors &waveFieldEvaluation, inteOptions opt);

  void evaluateIntegralOperatorGInverse(const Vectors &NonlinearPointWave,
                                        const Vector &nonlinearPointMat,
                                        Vectors &WaveFieldEvaluation,
                                        inteOptions opt);

  void
  evaluateIntegralOperator(const Vectors &NonlinearPointWave, const Vector &nonlinearPointMat,
                           Vectors &WaveFieldEvaluation, inteOptions opt);

  void evaluateIntegralOperatorGInverse(const FWIDataContainerDomain &nonlinearPoint,
                                        Vectors &waveFieldEvaluation, inteOptions opt);

  void evaluateDifferentialOperatorRK(const FWIDataContainerDomain &nonlinearPoint,
                                      Vectors &waveFieldEvaluation, inteOptions opt);

  void evaluateDifferentialOperatorRK(const Vector &Mat, const Vectors &waveFieldPoint,
                                      Vectors &waveFieldEvaluation, inteOptions opt);

  void
  evaluateAndAddLinearizedIntegralOperatorGInverse(const FWIDataContainerDomain &nonlinearPoint,
                                                   const Vector &matDirection,
                                                   Vectors &waveFieldEvaluation);

  void
  evaluateAndAddLinearizedIntegralOperatorGInverse(const Vectors &WaveFieldPoint,
                                                   const Vector &MatPoint,
                                                   const Vector &MatDirection,
                                                   Vectors &WaveFieldEvaluation);

  void evaluateLinearizedIntegralOperator(const FWIDataContainerDomain &nonlinearPoint,
                                          const Vector &matDirection,
                                          Vectors &LinearizedEvaluation);

  void evaluateLinearizedIntegralOperator(const Vectors &WaveFieldPoint, const Vector &MatPoint,
                                          const Vector &MatDirection, Vectors &WaveFieldEvaluation);

  void calculateWeakFormVectors(Vectors &b) const;

  void setAdjointFromRHS(const FWISeismogram &Seis, Vectors &U) const;


  void integrate(const Vectors &WaveFieldPoint, Vector &integratedVec,
                 int evaluationStep) const;


  void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                         const FWISubproblem &adjointLinearizedPoint,
                         FWIDataContainerDomain &adjointLinearizedEvaluation) override;

  void calculateUpdateIntegral(const Vectors &WaveFieldPoint, const Vector &Mat,
                               const Vectors &Lambda, Vector &AdjointMat,
                               Vectors *integratedVecs = nullptr,
                               const RKOperator *RkOp = nullptr) const;


  void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                  const FWIDataContainerDomain &linearizedPoint,
                  FWISubproblem &linearizedEvaluation) override;


  double
  scalarProductL2(const FWIDataContainerRange &u, const FWIDataContainerRange &v) const override;

  auto L2TScalarAAO(const Vectors &u, const Vectors &v) const -> double;

  auto L2tScalarAAO(const Vector &u, const Vector &v) const -> double;

  double
  scalarProductL2(const FWIDataContainerDomain &u, const FWIDataContainerDomain &v) const override;

  void integrateRKsolution(Vectors &WaveFieldEvaluation, bool fwd);

  void integrateRKsolutionFwd(Vectors &WaveFieldEvaluation);

  void integrateRKsolutionBwd(Vectors &WaveFieldEvaluation);

  RKOperator *getRKOperator(const Vector &Mat, const Vectors &WaveFieldPoint, inteOptions opt);

  void setSolutionFromRHSDeltaVADiscrete(const FWISeismogram &Seis, Vectors &u) const;

  void AdjointMeasurementDiscrete(const FWIDataContainerRange &adjointLinearizedPoint,
                                  FWIDataContainerDomain &adjointLinearizedEvaluation);

  void
  AdjointMeasurementDiscrete(const FWISeismogram &Seis, Vectors &AdjointLinearizedEvaluationWF);

  void assembleMeasurementTMeasurement(Matrix &M);

  void initializePrecon(const Vector &m, const Vector &m_update,
                        const Vectors &u_m) override;

  void initializePrecon(const FWIDataContainerDomain &m,
                        const FWIDataContainerDomain &m_update,
                        const FWISubproblem &u_m) override;


  void applyPrecon(const FWIDataContainerDomain &m,
                   FWIDataContainerDomain &m_update) override;

  void mSDProcedure(const FWIDataContainerDomain &iterateNonlinear,
                    FWIDataContainerDomain &iterateUpdatePN,
                    FWIDataContainerDomain &iterateMUpdate,
                    FWISubproblem &defectLinearized) override;


  void RestrictToVectorComp(Vector &u, int comp, int paraComps) const;

  void SetPenalty(const Vector &M_Upd, FWISubproblem &Defect) override;


  void Linearized(const Vectors &nonlinearPointWave, const Vector &nonlinearPointMat,
                  const Vectors &linearizedWave,
                  const Vector &linearizedMat, Vectors &linearizedEvaluationWave,
                  FWISeismogram &linearizedEvaluationSeismo);

  void GetInitialValueForInner(const FWIDataContainerDomain &iterateNonlinear,
                               const FWISubproblem &nonlinearEvaluation,
                               FWISubproblem &defectLinearized,
                               FWIDataContainerDomain &iterateMUpdate,
                               std::vector<int> prob) override;

  void assembleAAOInv(const std::vector<Vectors> &Evals, Matrix &PHess) const;

  void getLinearizedVectors(const Vector &m, const Vectors &u_fd, const Vector &Direc,
                            std::vector<Vectors> &Evals);
};

#endif
