#include "FWIForwardModelDGTSVAFullWave.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"
#include "FWITimeSolver.hpp"
#include "Reduced/FWILinOpRed.hpp"
#include "Assembler/FWIRhsFromVectorsVA.hpp"

void FWIForwardModelDGTSVAFullWave::NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                                                        FWISubproblem &subproblemEvaluation,
                                                        std::vector<int> subproblem) {
    for (size_t i = 0; i < subproblemEvaluation.size(); ++i) {
        size_t shot = subproblem[i];
        problem_fwd->setShot(shot);
        Parameter2SolutionMap(nonlinearPoint.getMat(),
                              subproblemEvaluation[i].getVectorsWrite());
        uMFull[i] = &(subproblemEvaluation[i].getVectors());
    }
    problem_fwd->setShot(-1);
}


void FWIForwardModelDGTSVAFullWave::calculateDefect(const std::vector<FWIDataContainerRange> &data_obs, const FWISubproblem &nonlinearEvaluation, FWISubproblem &defect,std::vector<int> subproblem) {
    for (size_t i =0; i < subproblem.size();++i){
        defect[i].getVectorsWrite() =  data_obs[subproblem[i]].getVectors();
        defect[i].getVectorsWrite() -= nonlinearEvaluation[i].getVectors();
    }
}


void FWIForwardModelDGTSVAFullWave::AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint, const FWISubproblem &adjointLinearizedPoint, FWIDataContainerDomain &adjointLinearizedEvaluation) {
    FWIDataContainerDomain tmpMat(adjointLinearizedEvaluation);
    for (int i =0; i < uMFull.size(); ++ i ){

        std::unique_ptr<FWIRhs> rhsFromFull = std::make_unique<FWIRhsFromVectorsVA>(
                adjointLinearizedPoint[i].getVectors(), eleCacheContainer->Get(), TConf.dt);
        AdjointLinearizedParameter2SolutionMap(nonlinearPoint.getMat(), *(uMFull[i]),
                                               *rhsFromFull, nullptr,
                                               tmpMat.getMatWrite());
        adjointLinearizedEvaluation.getMatWrite() += tmpMat.getMat();
    }
}


void FWIForwardModelDGTSVAFullWave::Linearized(const FWIDataContainerDomain &nonlinearPoint, const FWIDataContainerDomain &linearizedPoint, FWISubproblem &subprob) {
    FWIDataContainerRange tmp(getReconstructionType(), GetWaveVec(), TConf.timeSteps);
    for (int i =  0; i< uMFull.size(); ++ i){
        LinearizedParameter2SolutionMap(nonlinearPoint.getMat(),linearizedPoint.getMat(),
                                        *(uMFull[i]),&tmp.getVectorsWrite());
        subprob[i].getVectorsWrite() += tmp.getVectors();
    }

}

void FWIForwardModelDGTSVAFullWave::LinearizedParameter2SolutionMap(const Vector &m, const Vector &m_test,
                                                                    const Vectors &u_m, Vectors *u_linearized) {
    FWILinOpRed LinOp(u_m, m, problem_fwd, PMLH, TConf, eleCacheContainer->Get(), false);
    LinOp.Solve(m_test, *u_linearized);
}

double
FWIForwardModelDGTSVAFullWave::scalarProductL2(const FWIDataContainerRange &u, const FWIDataContainerRange &v) const {
    return L2TScalar(u.getVectors(), v.getVectors());
}

void FWIForwardModelDGTSVAFullWave::Parameter2SolutionMap(const Vector &m, Vectors &u_m) {
    std::shared_ptr<DGAssemble> assemble = getDGAssemble(m, *problem_fwd, *problem_fwd, true);
    FWITimeSolver TSolveR(*FWIDiscStack.wV, PMLH, TConf.rkorder);
    TSolveR.Solve(assemble, u_m);
}

double
FWIForwardModelDGTSVAFullWave::scalarProductL2(const FWIDataContainerDomain &u, const FWIDataContainerDomain &v) const {
    return L2Scalar(u.getMat(), v.getMat());
}

void FWIForwardModelDGTSVAFullWave::AdjointLinearizedParameter2SolutionMap(const Vector &m, const Vectors &u_m,
                                                                           const FWIRhs &v, Vectors *lambda,
                                                                           Vector &adjoint_m_mv) {
    auto AdjointRHS = std::make_shared<DGVAcousticRHSAssemblerAdjoint>(*Obs);
    auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem_fwd, m, PMLH.get());
    auto matrixAssembler = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), false);
    auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssembler, AdjointRHS);
    calculateAdjointWithoutSaving(m, assembleCombined, adjoint_m_mv);
}

void FWIForwardModelDGTSVAFullWave::calculateAdjointWithoutSaving(const Vector &Material,
                                                                  const std::shared_ptr<ILinearTimeAssemble> &assemble,
                                                                  Vector &Adjoint_m_mv) const {
    std::unique_ptr<TimeIntegrator> timeIntegrator = TimeIntegratorCreator().
            WithLinearSolver(GetLinearSolver("LS", GetPC("PointBlockJacobi"))).
            WithLinearSolver(GetLinearSolver("LS", GetPC("PointBlockJacobi"))).
            WithLinearSolver(GetLinearSolver("LS", GetPC("PointBlockJacobi"))).
            WithLinearSolver(GetLinearSolver("LS", GetPC("PointBlockJacobi"))).
            WithConfigTypeEntry().
            CreateUnique();
    std::unique_ptr<Vector> LambdaStep2;
    std::unique_ptr<Vector> LambdaStepUnpadded;
    if (PMLH->isPML()) {
        LambdaStep2 = std::make_unique<PMLWaveVectorFWI>(*PMLH);
        LambdaStepUnpadded = std::make_unique<Vector>((*uM[0])[0]);
        LambdaStepUnpadded->Clear();
    } else {
        LambdaStep2 = std::make_unique<Vector>((*uM[0])[0]);
    }
    LambdaStep2->Clear();
    Adjoint_m_mv.Clear();
    timeIntegrator->Initialize(assemble.get(), *LambdaStep2);
    timeIntegrator->SetDivergenceCheck(false);
    mout.StartBlock("TI");
    mout << "Starting " << timeIntegrator->Name() << " (Order=" << timeIntegrator->GetOrder() << ")" << endl;
    while (!assemble->IsFinished()) {
        timeIntegrator->OneTimeStep(assemble.get(), *LambdaStep2);
        if (PMLH->isPML()) {
            (*LambdaStepUnpadded).Clear();
            PMLWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *LambdaStep2, *LambdaStepUnpadded);
            calculateL2AdjointStep(Material, (*uM[0]), assemble->Step(), *LambdaStepUnpadded, Adjoint_m_mv);
        } else {
            calculateL2AdjointStep(Material, (*uM[0]), assemble->Step(), *LambdaStep2, Adjoint_m_mv);
        }
    }
    timeIntegrator->Destruct();
    mout.EndBlock();
    Adjoint_m_mv *= -1.0;
}

void
FWIForwardModelDGTSVAFullWave::calculateL2AdjointStep(const Vector &Material, const Vectors &U_m, int step,
                                                      const Vector &LambdaStep,
                                                      Vector &Adjoint_m_mv) const {
    int StepBwd = (U_m).size() - 1 - step;
    const Vector &U_m_bwd(U_m[StepBwd]);
    Vector U_m_fd(U_m[StepBwd + 1]);
    U_m_fd -= U_m_bwd;
    U_m_fd *= (1.0 / TConf.dt);
    FWIAdjointCalculator fwiAdjointCalculator(Material, problem_fwd->getParametrization(), eleCacheContainer->Get(), recon,
                                              numL, TConf.dt, problem_fwd);
    fwiAdjointCalculator.calculateL2AdjointOneStep(U_m_bwd, U_m_fd, LambdaStep, Adjoint_m_mv);
}