#include "FWISeismogram.hpp"
#include "../../../DSPFilters/Filter.h"
#include "../../../DSPFilters/Dsp.h"
#include <algorithm>
#include <memory>
#include <fftw3.h>
#include <filesystem>

//std::ofstream& operator << (std::ofstream& o, const FWIReceiverTimeseries& ts) {
//        o << std::to_string(static_cast<int>(ts.size())).c_str() << endl;
//        for (auto m=ts.begin(); m!=ts.end(); ++m) {
//            o << "  " << m->first << '\t';
//            const FWIState& state = *m->second;
//            for (int i=0; i<ts.state_comps; ++i)
//                o << std::to_string(static_cast<double>(state[i])).c_str() << " ";
////            o << *m->second;
//            o << endl;
//        }
//        return o;
//    }

//    friend std::istream& operator >> (std::istream& in, FWIReceiverTimeseries& ts) {
//        ts.clear();
//        int ts_size = 0;
//        in >> ts_size;
//        for (int i=0; i<ts_size; ++i) {
//            double t;
//            FWIState* state = new FWIState(0.0, ts.state_comps);
//            in >> t;
//            for (int i=0; i<ts.state_comps; ++i)
//                in >> (*state)[i];
////            in >> *state;
//            ts.push_back(std::make_pair(t, state));
//        }
//        return in;
//    }


FWISeismogram::FWISeismogram() : state_comps(0) {}

FWISeismogram::FWISeismogram(const FWIObservationSpecification &spec, int state_comps)
        : spec(spec), state_comps(state_comps) {
    for (auto &re: spec) {
        for (int i = 0; i < spec.nSteps(); ++i) {
            double t = spec.step(i);
            addMeasurement(re, t, FWIState(0.0, state_comps));
        }
    }
}

FWISeismogram::FWISeismogram(const char *filename, int state_comps, const std::unordered_map<Point, std::list<Point>>&
ReceiverList)
        : state_comps(state_comps) {
    readFromFile(filename);
  for (auto & re : spec) {
    if (!ReceiverList.contains(re)) {
      timeseries.erase(re);
      }
  }
}

FWISeismogram::FWISeismogram(const FWISeismogram &s)
        : spec(s.spec), state_comps(s.state_comps) {
    for (auto ts = s.ts_begin(); ts != s.ts_end(); ++ts)
        timeseries[ts->first] = new FWIReceiverTimeseries(*ts->second);
}
FWISeismogram::~FWISeismogram() { ts_clear(); }

//    double alpha() const { return spec.timeDelta()*spec.timeDelta(); }
double FWISeismogram::alpha() const { return spec.timeDelta(); }

//double FWISeismogram::alpha() const { return 1.0; }
void FWISeismogram::LowPassFilter(double freq, int order) {
    if (state_comps > 1) {
        THROW("Only for one state component defined!");
    }
    int numSamples = spec.nSteps();
    double *seismoData[1];
    seismoData[0] = new double[numSamples];
    double *data[1];
    data[0] = new double[numSamples];
    for (auto &t: timeseries) {
      auto f = std::make_unique<Dsp::SimpleFilter<Dsp::Butterworth::LowPass<4>, 1>>();
        f->setup(order, 1.0 / spec.timeDelta(), freq);
        auto &ts = *t.second;
        for (int i = 0; i < ts.size(); ++i) {
            seismoData[0][i] = (*ts[i].second)[0];
        }
        f->process(numSamples, seismoData);
        for (int i = 0; i < ts.size(); ++i) {
            data[0][i] = seismoData[0][ts.size() - 1 - i];
        }
        f->process(numSamples, data);
        for (int i = 0; i < ts.size(); ++i) {
            (*ts[ts.size() - 1 - i].second)[0] = data[0][i];
        }
    }
    delete[] data[0];
    delete[] seismoData[0];
}
void FWISeismogram::BandPassFilter(double lowerFreq, double upperFreq, int order) {
  if (state_comps > 1) {
    THROW("Only for one state component defined!");
  }
  double centerFreq = (upperFreq + lowerFreq)/ 2.0;
  double width = upperFreq - centerFreq;
  int numSamples = spec.nSteps();
  double *seismoData[1];
  seismoData[0] = new double[numSamples];
  double *data[1];
  data[0] = new double[numSamples];
  for (auto &t: timeseries) {
    auto f = std::make_unique<Dsp::SimpleFilter<Dsp::Butterworth::BandPass<4>, 1>>();
    f->setup(order, 1.0 / spec.timeDelta(), centerFreq, width);
    auto &ts = *t.second;
    for (int i = 0; i < ts.size(); ++i) {
      seismoData[0][i] = (*ts[i].second)[0];
    }
    f->process(numSamples, seismoData);
    for (int i = 0; i < ts.size(); ++i) {
      data[0][i] = seismoData[0][ts.size() - 1 - i];
    }
    f->process(numSamples, data);
    for (int i = 0; i < ts.size(); ++i) {
      (*ts[ts.size() - 1 - i].second)[0] = data[0][i];
    }
  }
  delete[] data[0];
  delete[] seismoData[0];
}
void FWISeismogram::BandPassFilterIFOS(double lowerFreq, double upperFreq) {
  double F_HP = lowerFreq;
  double fc = upperFreq;
  int NT = spec.nSteps();
  int Npad = (int)(pow(2.0, ceil(log((double)(NT))/log(2.0))+1.0) );
  double df = (double) 1.0/(Npad)/spec.timeDelta();
  vector<double> ff(Npad+1),BFabs(Npad+1);
  int j =0;
  for (j=0;j<Npad/2;j++){
    ff[j+1] = (double) j * df;
  }
  for (j=Npad/2;j<Npad;j++){
    ff[j+1] = -(double) (Npad-j) * df;
  }
  int it1=floor(1/df);
  int it2=round(F_HP/df);
  int it3=floor(fc/df);
  int it4 = 2*it3;
  int flank1 = it2 - it1;
  int flank2 = it4-it3;
  if(PPM->Proc()==0) {
    printf("    Band-pass corner indexes    : it1 = %d, it2 = %d, it3 = %d, it4 = %d \n",
           it1,it2,it3,it4);
    printf("    Band-pass corner frequencies: Fc1 = %.2f, Fc2 = %.2f, Fc3 = %.2f, Fc4 = %.2f \n",
           ff[it1],ff[it2],ff[it3],ff[it4]);
  }
  BFabs[0] = 0.0;
  for (int n=0; n<=Npad; n++) {
    if ((n>=it1) && (n<it3))	BFabs[n] = 1.0;
    if ((n>=it3) && (n<it4))	BFabs[n] = (double) (1.0-(1-cos((double)(n-it3)/flank2*M_PI))/2.0);
    if(n>=it4) BFabs[n] = 0.0; // BFabs[n];
    if(F_HP) {
      if (n<=it1) BFabs[n] = 0.0;
      if ((n>it1) && (n<it2))	BFabs[n] = (double) ((1.0-cos((double)(n-it1)/flank1*M_PI))/2.0);
      if ((n>=it2) && (n<it3))	BFabs[n] = 1.0;
    }
  }
  for (auto &t: timeseries) {
    std::vector<std::complex<double>> in_data(Npad,0.0);
    std::vector<std::complex<double>> out_data(Npad,0.0);

    auto &ts = *t.second;
    if (ts.size() == 0) continue;
    for (int i = 0; i < ts.size()-1; ++i) {
      in_data[i].real((*ts[i+1].second)[0]);
      in_data[i].imag(0.0);
      out_data[i].real(0.0);
      out_data[i].imag(0.0);
    }
    for (j = ts.size()-1; j < Npad; j++) {
      in_data[j].real(0.0);
      in_data[j].imag(0.0);
      out_data[j].real(0.0);
      out_data[j].imag(0.0);
    }
    fftw_plan p_data;
    p_data = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(in_data.data()),
                              reinterpret_cast<fftw_complex *>(out_data.data()), 1, FFTW_ESTIMATE);
    fftw_execute(p_data);
    for (j = 1; j <= Npad / 2; j++) {
      out_data[j] = out_data[j] * BFabs[j];
      if(j!= 1) {
        out_data[Npad + 1 - j] = conj(out_data[j]) * BFabs[j];
      }
      in_data[j].imag(0.0);
      in_data[j].real(0.0);
    }
    out_data[0].real(0.0);
    out_data[0].imag(0.0);
    out_data[1].real(0.0);
    out_data[1].imag(0.0);
    fftw_destroy_plan(p_data);
    fftw_plan p_data2;

    p_data2 = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(out_data.data()),
                              reinterpret_cast<fftw_complex *>(in_data.data()), -1, FFTW_ESTIMATE);
    fftw_execute(p_data2);
    for (int i = 1; i < ts.size(); ++i) {
      (*ts[i].second)[0] = in_data[i-1].real()/Npad;
    }
    fftw_destroy_plan(p_data2);
  }
}
FWISeismogram& FWISeismogram::operator = (const FWISeismogram& s) {
    spec = s.spec;
    state_comps = s.state_comps;

    ts_clear();
    for (auto ts=s.ts_begin(); ts!=s.ts_end(); ++ts) {
        timeseries[ts->first] = new FWIReceiverTimeseries(*ts->second);
    }
    return *this;
}
void FWISeismogram::setSpecification(const FWIObservationSpecification& spec_new) {
    ts_clear();
    spec = spec_new;
    for (auto r=spec.begin(); r!=spec.end(); ++r)
        timeseries[*r] = new FWIReceiverTimeseries(state_comps);
}
int FWISeismogram::getStateComponents() const { return state_comps; } // TODO: sollte auch in Datei gespeichert werden
void FWISeismogram::setStateComponents(int state_comps_new) {
    state_comps = state_comps_new;
    for (auto ts=ts_begin(); ts!=ts_end(); ++ts)
        ts->second->setStateComponents(state_comps_new);
}

void FWISeismogram::addMeasurement(const Point& receiver, double t, const FWIState& measurement) {
    addMeasurement(receiver, t, new FWIState(measurement));
    //        if (timeseries.find(receiver) == timeseries.end())
    //            timeseries.emplace(std::std::make_pair(receiver, new FWIReceiverTimeseries(*assemble)));
    //        timeseries.at(receiver).push_back(std::make_pair(t, new FWIState(measurement)));
}
void FWISeismogram::addMeasurement(const Point& receiver, double t, FWIState* measurement) {
    if (timeseries.find(receiver) == timeseries.end())
        timeseries[receiver] = new FWIReceiverTimeseries(state_comps);
    timeseries[receiver]->push_back(std::make_pair(t, measurement));
}

const FWIReceiverTimeseries &FWISeismogram::getTimeSeries(const Point &receiver) const {
    return *timeseries.at(receiver);
}

FWIReceiverTimeseries &FWISeismogram::getTimeSeriesWrite(const Point &receiver) {
    return *timeseries.at(receiver);
}

const FWIObservationSpecification &FWISeismogram::getSpecification() const { return spec; }
bool FWISeismogram::findRelevantProc(std::unordered_map<Point, FWIReceiverTimeseries*>::iterator ts) {
    for (auto &i: *ts->second) {
        const FWIState &state = *i.second;
        for (double k: state) {
            if (abs(k) > 0) {
                return false;
            }
        }
    }
    return true;
}
void FWISeismogram::collect(bool average) {
    Date Start;
    ExchangeBuffer E;
    MOUT(state_comps);
    if (!PPM->master()) {
        for (auto ts = timeseries.begin(); ts != timeseries.end(); ++ts) {
            bool zero2 = true;
            zero2 = findRelevantProc(ts);
            if (!zero2) {
                E.Send(0) << static_cast<Point>(ts->first);
                E.Send(0) << static_cast<int> (ts->second->size());
                for (auto i = ts->second->begin(); i != ts->second->end(); ++i) {
                    const FWIState &state = *i->second;
                    E.Send(0) << static_cast<double>(i->first);
                    for (unsigned int k = 0; k < state.size(); ++k)
                        E.Send(0) << state[k];
                }
            }
        }
    }
    E.Communicate();
//    mout << "Time needed for measuring til E.Communicate: " <<  Date() - Start << endl;
//    MOUT("FAIL");
    if (PPM->master()) { // kommunikation auf den master, da auch auf Kanten gemessen werden kann (und wird) gibt es für einzelne receiver doppelte werte
        for (short q=0; q<PPM->size(); ++q) {
            while (E.Receive(q).size() < E.ReceiveSize(q)) {
                Point receiver;
                E.Receive(q) >> receiver;
                int ts_size;
                E.Receive(q) >> ts_size;
                if (timeseries.find(receiver) == timeseries.end()) {
                    timeseries[receiver] = new FWIReceiverTimeseries(state_comps);
                }
                FWIReceiverTimeseries* current_ts = timeseries[receiver];
                for (int i=0; i<ts_size; ++i) {
                    double t;
                    E.Receive(q) >> t;
                    auto *state = new FWIState(0.0, state_comps);
                    for (unsigned int k=0; k<state->size(); ++k)
                        E.Receive(q) >> (*state)[k];
                    current_ts->push_back(std::make_pair(t,state));
                }
            }
        }
//        mout << "Time needed for first for loop: " <<  Date() - Start << endl;

        for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
            FWIReceiverTimeseries* current_ts = ts->second;

           std::vector< std::pair<double,FWIState*> > tmp(current_ts->begin(), current_ts->end());
            std::sort(tmp.begin(), tmp.end(), comp_timeseries_entry);
            current_ts->clear();
            for (auto e=tmp.begin(); e!=tmp.end(); ++e) { // mehrere Messungen auf verschiedenen cores werden addiert und gemittelt(oder nicht)
                auto e1=e;
                FWIState *sum = e->second;
                ++e1;
                int counter = 1;
                while(e1 != tmp.end() && e1->first == e->first) {
                    *sum += *e1->second; // es muss ausgedünnt werden. Die doppelten keys werden gelöscht
                    delete e1->second;
                    ++counter;
                    ++e1;
                }
                if (average)
                    *sum /= counter;

                current_ts->push_back(std::make_pair(e->first,sum));
                e += (counter-1);
            }
        }
    }
//    mout << "Time needed for the rest: " <<  Date() - Start << endl;

}
void FWISeismogram::broadcast() {
    if (PPM->master()) {
        PPM->Broadcast((int)timeseries.size());
        for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
            PPM->Broadcast((Point)ts->first);
            PPM->Broadcast((int) ts->second->size());
            for (auto m=ts->second->begin(); m!=ts->second->end(); ++m) {
                PPM->Broadcast((double) m->first);

                for (int k=0; k<state_comps; ++k)
                    PPM->Broadcast((*m->second)[k]);
            }
        }
    }
    else {
        timeseries.clear();
        int ts_count;
        PPM->Broadcast(ts_count);
        for (int ts=0; ts<ts_count; ++ts) {
            Point receiver;
            PPM->Broadcast(receiver);
            int ts_size;
            PPM->Broadcast(ts_size);

            if (timeseries.find(receiver) == timeseries.end())
                timeseries[receiver] = new FWIReceiverTimeseries(state_comps);

            FWIReceiverTimeseries* current_ts = timeseries.at(receiver);
            for (int i=0; i<ts_size; ++i) {
                double t;
                PPM->Broadcast(t);

                FWIState* m = new FWIState(0.0, state_comps);
                for (int k=0; k<state_comps; ++k)
                    PPM->Broadcast((*m)[k]);

                current_ts->push_back(std::make_pair(t,m));
            }
        }
    }
}

void FWISeismogram::operator = (double val) {
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        for (auto m=current_ts->begin(); m!=current_ts->end(); ++m)
            *m->second = val;
    }
}

void FWISeismogram::operator *= (double val) {
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        for (auto m=current_ts->begin(); m!=current_ts->end(); ++m)
            *m->second *= val;
    }
}
void FWISeismogram::fillSeismogramWithNormalGaussianNoise(FWISeismogram &noiseSeismo, unsigned int seed) {
  noiseSeismo =0.0;
  std::random_device rd;
  unsigned int localSeed = seed == 0 ? rd() : seed;
  unsigned int rnd = localSeed+PPM->Proc()*1000;
  std::normal_distribution<double> distribution(0.0, 1.0);
  for (auto ts=noiseSeismo.ts_begin(); ts!=timeseries.end(); ++ts) {
    if (ts->second->empty()) continue;
    std::mt19937 generator(rnd++);
    for (auto [_,noiseMeasurement] : *ts->second){
      noiseMeasurement[0] = distribution(generator);
    }
  }
  double noisePower = noiseSeismo*noiseSeismo;
  noiseSeismo *= 1.0/sqrt(noisePower);
}

void FWISeismogram::addRelativeNoise(double signalToNoiseRatio, unsigned int seed) {
    auto noiseSeismo = FWISeismogram(*this);
  fillSeismogramWithNormalGaussianNoise(noiseSeismo, seed);
    double signalPower = sqrt((*this)*(*this));
    noiseSeismo *=  signalPower/ signalToNoiseRatio;
    (*this) += noiseSeismo;
}

void FWISeismogram::setComponent (int comp, double val) {
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        for (auto m=current_ts->begin(); m!=current_ts->end(); ++m)
            (*m->second)[comp] = val;
    }
}

void FWISeismogram::operator -= (const FWISeismogram& s) {
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        const FWIReceiverTimeseries& current_tss = s.getTimeSeries(ts->first);
        auto ms = current_tss.begin();
        for (auto m=current_ts->begin();
             m!=current_ts->end() && ms!=current_tss.end(); ++m, ++ms) {
            *m->second -= *ms->second;
        }
    }
}
void FWISeismogram::operator += (const FWISeismogram& s) {
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        const FWIReceiverTimeseries& current_tss = s.getTimeSeries(ts->first);
        auto ms = current_tss.begin();
        for (auto m=current_ts->begin();
             m!=current_ts->end() && ms!=current_tss.end(); ++m, ++ms) {
            *m->second += *ms->second;
        }
    }
}

std::unordered_map<Point, FWIReceiverTimeseries*>::iterator FWISeismogram::ts_begin() { return timeseries.begin(); }
std::unordered_map<Point, FWIReceiverTimeseries*>::iterator FWISeismogram::ts_end() { return timeseries.end(); }
void FWISeismogram::ts_clear() {
    if (state_comps != 0) {
        for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts){
            delete ts->second;
        }
    }
    timeseries.clear();
}

std::unordered_map<Point, FWIReceiverTimeseries*>::const_iterator FWISeismogram::ts_begin() const { return timeseries.begin(); }
std::unordered_map<Point, FWIReceiverTimeseries*>::const_iterator FWISeismogram::ts_end() const { return timeseries.end(); }

// -- IO stuff --
void FWISeismogram::printSpecification() const { spec.printInfo(); }
void FWISeismogram::printObservation() const {
    mout << "FWIObservationRecording:" << endl;
    spec.printInfo("  ");
    mout << "  measurements:" << endl;
    for (auto ts=timeseries.begin(); ts!=timeseries.end(); ++ts) {
        FWIReceiverTimeseries* current_ts = ts->second;
        mout << "    receiver: " << ts->first << endl;
        for (auto m=current_ts->begin(); m!=current_ts->end(); ++m)
            mout << "      t: " << std::setw(6) << m->first << "  -  " << *m->second << endl;
    }
}

void FWISeismogram::readFromFile(const std::string& filename) {
    std::ifstream file;
    file.open(filename.c_str());
    if (file.fail()) {
        Exit("Could not read seismogram with the filename: " + filename);
    }
    file >> *this;
    file.close();

}

void FWISeismogram::writeToFile(const std::string &filename) const {
    if (PPM->proc() != 0) return;
    std::ofstream file;
  file.open(filename.c_str());

  if (file.fail()) {
    if (!std::filesystem::is_directory(filename.substr(0,filename.rfind('/')))) {
      std::filesystem::create_directory(filename.substr(0,filename.rfind('/')));
    }
    file.open(filename.c_str());
  }

    if (file.fail()) {
        Exit("Could not write seismogram");
    }
    file << *this;
    file.close();
}

//double operator*(const FWISeismogram &s1, const FWISeismogram &s2) {
//    double res = 0.0;
//    const double alpha = s1.alpha();
//    for (auto ts1 = s1.ts_begin(); ts1 != s1.ts_end(); ++ts1) {
//        FWIReceiverTimeseries *current_ts1 = ts1->second;
//        const FWIReceiverTimeseries &current_ts2 = s2.getTimeSeries(ts1->first);
//        for (int i = 0; i < current_ts1->size(); ++i) {
//            auto val = (*(*current_ts1)[i].second) * (*(current_ts2)[i].second);
//            if (i == 0 || i == current_ts1->size() - 1) val *= 0.5;
//            res += alpha * val;
//        }
//    }
//    return PPM->Sum(res); // VERRY WEIRD IMPLEMENTATION OF SEISMOGRAMS NOW! MASTER IS RESPONSIBLE FOR
//
//}
double operator*(const FWISeismogram &s1, const FWISeismogram &s2) {
  double res = 0.0;
  for (auto ts1 = s1.ts_begin(); ts1 != s1.ts_end(); ++ts1) {
    if(ts1->second->size()>0 && s2.getTimeSeries(ts1->first).size() > 0) {
      res += (*ts1->second) * s2.getTimeSeries(ts1->first);
    }
  }
  return PPM->Sum(res); // VERRY WEIRD IMPLEMENTATION OF SEISMOGRAMS NOW! MASTER IS RESPONSIBLE FOR

}

void NormalizeTracewise(FWISeismogram &Seis) {
  for (auto rec: Seis.getSpecification()){
    auto &ts = Seis.getTimeSeriesWrite(rec);
    double norm = sqrt(ts * ts);
    ts *= 1.0 / norm;
  }
}

void DivideByFactorTracewise(FWISeismogram &Seis, const std::unordered_map<Point, double> &factors) {
  for (auto rec: Seis.getSpecification()){
    auto &ts = Seis.getTimeSeriesWrite(rec);
    ts *= 1.0 / factors.at(rec);
  }
}

void
MultiplyByFactorTracewise(FWISeismogram &Seis, const std::unordered_map<Point, double> &factors) {
  for (auto rec: Seis.getSpecification()){
    auto &ts = Seis.getTimeSeriesWrite(rec);
    ts *= factors.at(rec);
  }
}

void TraceWiseScalarProduct(const FWISeismogram &Seis1, const FWISeismogram &Seis2,
                            std::unordered_map<Point, double> &factors) {
  for (auto rec: Seis1.getSpecification()){
    auto &ts1 = Seis1.getTimeSeries(rec);
    auto &ts2 = Seis2.getTimeSeries(rec);
    factors[rec] = ts1 * ts2;
  }
}

CMatrix FWISeismogram::GetFFT() const {
  if (state_comps > 1) {
    THROW("Only for one state component defined!");
  }
  int numSamples = spec.nSteps();
  CMatrix FFTMat(0.0,timeseries.size(),numSamples);
  RVector SeismoDataInput(0.0,spec.nSteps());
  CVector SeismoDataOutput(0.0,spec.nSteps());
  int j = 0;
  for (auto &t: timeseries) {
    auto &ts = *t.second;
    for (int i = 0; i < ts.size(); ++i) {
      SeismoDataInput[i] = (*ts[i].second)[0];
    }
    FFT::RealToComplexVector(SeismoDataInput,SeismoDataOutput);
    for (int i = 0; i < ts.size(); ++i) {
      FFTMat[j][i] = SeismoDataOutput[i];
    }
    ++j;
    SeismoDataOutput = 0.0;
  }
  return FFTMat;
}

