#ifndef _FWISEISMOGRAM_H_
#define _FWISEISMOGRAM_H_

#include <random>
#include <fstream>
#include <iomanip>
#include "TimeDate.hpp"
#include "m++.hpp"
#include "Point.hpp"
#include "RVector.hpp"
#include "Parallel.hpp"
#include "CMatrix.hpp"
#include "FFT.hpp"
#include "IFOSToFWI.hpp"
#include "Observation/FWIObservationSpecification.hpp"
#include "FWIReceiverTimeseries.hpp"
#include "segy.h"
#include <iostream>
#define SHIFT_IND 1
#define FREE_ARGUMENT char *

inline bool comp_timeseries_entry(const std::pair<double, FWIState *> &a,
                                  const std::pair<double, FWIState *> &b) {
  return (a.first < b.first);
}


class FWISeismogram {
protected:
  FWIObservationSpecification spec;
  std::unordered_map<Point, FWIReceiverTimeseries *> timeseries;
  int state_comps;
public:
  FWISeismogram();

  FWISeismogram(const FWIObservationSpecification &spec, int state_comps);

  FWISeismogram(const char *filename, int state_comps, const std::unordered_map<Point, std::list<Point>>&
  ReceiverList);

  FWISeismogram(const FWIObservationSpecification &spec, const std::unordered_map<Point, std::list<Point>>&
                ReceiverList, const char *filename_data) : spec(spec),
  state_comps(1){
    readIFOSshot(ReceiverList,filename_data);
  }

  FWISeismogram(const FWISeismogram &s);

  ~FWISeismogram();

//    double alpha() const { return spec.timeDelta(); }
  double alpha() const;

  void LowPassFilter(double, int);

    void BandPassFilter(double lowerFreq, double UpperFreq, int order);
  void BandPassFilterIFOS(double lowerFreq, double upperFreq);
//  void BandPassFilter(double lowerFreq, double UpperFreq, int order);
    CMatrix GetFFT() const;

  void fillSeismogramWithNormalGaussianNoise(FWISeismogram &noiseSeismo, unsigned int seed);

  FWISeismogram &operator=(const FWISeismogram &s);

  void setSpecification(const FWIObservationSpecification &spec_new);

  int getStateComponents() const; // TODO: sollte auch in Datei gespeichert werden

  void setStateComponents(int state_comps_new);

  void addMeasurement(const Point &receiver, double t, const FWIState &measurement);

  void addMeasurement(const Point &receiver, double t, FWIState *measurement);

  const FWIReceiverTimeseries &getTimeSeries(const Point &receiver) const;

  FWIReceiverTimeseries &getTimeSeriesWrite(const Point &receiver);

  const FWIObservationSpecification &getSpecification() const;

  bool findRelevantProc(std::unordered_map<Point, FWIReceiverTimeseries *>::iterator);

  void collect(bool average = false);

  void broadcast();

  void operator=(double val);

  void operator*=(double val);

  void addRelativeNoise(double signalToNoiseRatio, unsigned int seed);

  void setComponent(int comp, double val);

  void operator-=(const FWISeismogram &s);

  void operator+=(const FWISeismogram &s);

  friend double operator*(const FWISeismogram &s1, const FWISeismogram &s2);

  std::unordered_map<Point, FWIReceiverTimeseries *>::iterator ts_begin();

  std::unordered_map<Point, FWIReceiverTimeseries *>::iterator ts_end();

  void ts_clear();

  std::unordered_map<Point, FWIReceiverTimeseries *>::const_iterator ts_begin() const;

  std::unordered_map<Point, FWIReceiverTimeseries *>::const_iterator ts_end() const;

  // -- IO stuff --
  void printSpecification() const;

  void printObservation() const;

  void readFromFile(const std::string &filename);

  void writeToFile(const std::string &filename) const;

  friend std::ofstream &operator<<(std::ofstream &o, const FWISeismogram &rec) {
    o << std::fixed << std::setprecision(15);
    o << rec.spec << endl;
    o << rec.state_comps << endl;
    o << rec.timeseries.size() << endl;

    for (auto ts = rec.timeseries.begin(); ts != rec.timeseries.end(); ++ts) {
      const Point &Rec = ts->first;
      for (int d = 0; d < Rec.SpaceDim(); ++d)
        o << ts->first[d] << " ";
      o << endl;
//            o << ts->first << endl;
      o << *ts->second;
    }
    return o;
  }

  friend std::istream &operator>>(std::istream &in, FWISeismogram &rec) {
    rec.spec.clear();
    in >> rec.spec;
    in >> rec.state_comps;
    int ts_count = 0;
    in >> ts_count;
    for (int i = 0; i < ts_count; ++i) {
      Point p = Origin;
      for (int d = 0; d < p.SpaceDim(); ++d)
        in >> p[d];
      auto *current_ts = new FWIReceiverTimeseries(rec.state_comps);
      in >> *current_ts;
      rec.timeseries[p] = current_ts;
    }
    return in;
  }

  void readIFOSshot(const std::unordered_map<Point, std::list<Point>>& ReceiverList,const char *data) {
    ts_clear();
    int SamplingFactor = 1;
    Config::Get("SpecfileSamplingFactor",SamplingFactor);
    int nrSamples =6666;
    Config::Get("SpecfileNrOfSamples",nrSamples);

    int ntr_glob = spec.size();
    int NT = spec.nSteps()-1;
    segy tr;
    float dump;
    FILE *fpdata;
    mout << "READING: " << data << endl;
    fpdata = fopen(data, "r");
    int tracl1 = 1;
    for ( auto re =  spec.begin(); re != spec.end();re++, tracl1++) {
      fread(&tr, 240, 1, fpdata);
      fread(&tr.data, 4,  nrSamples, fpdata);
      if (!ReceiverList.contains(*re)) {
        addMeasurement(*re, 0.0, new FWIState(0.0, 1));
        for (int j = 1; j < NT + 1; j++) {
          addMeasurement(*re, spec.step(j), new FWIState(0.0, 1));
        }
      } else {
        addMeasurement(*re, 0.0, new FWIState(0.0, 1));
        for (int j = 1; j < NT+1; j++) {
           dump = tr.data[SamplingFactor*j-1];
//           MOUT(SamplingFactor*j - 1);
//           MOUT(spec.step(j));
          addMeasurement(*re, spec.step(j), new FWIState(dump, 1));
        }
      }
    }
    fclose(fpdata);
  }
};

void NormalizeTracewise(FWISeismogram &Seis);

void DivideByFactorTracewise(FWISeismogram &Seis,const  std::unordered_map<Point,double> &factors);

void MultiplyByFactorTracewise(FWISeismogram &Seis,const  std::unordered_map<Point,double> &factors);

void TraceWiseScalarProduct(const FWISeismogram &Seis1,
                            const FWISeismogram &Seis2,
                            std::unordered_map<Point,double> &factors);
#endif
