//
// Created by cr on 03.11.22.
//

#ifndef FWI_PMLHANDLER_HPP
#define FWI_PMLHANDLER_HPP

#include "FWIHelpers.hpp"
#include "DGDiscretization.hpp"
#include "MeshesCreator.hpp"
#include "DGAcousticElement.hpp"
class CosTaperConfig {
public:
  double deltad = 0.3;
  double vQ = 0.85;
  double vB = 0.2;
  double Qb = 1.1;
  double alpha1 = -1;
  double alpha2 = -2;
  CosTaperConfig(double alpha1_, double alpha2_, double deltad_ = 0.3,  double vQ_ = 0.85,
                double vB_ = 0.2,  double Qb_ = 1.1) : alpha1(alpha1_),alpha2(alpha2_),
                deltad(deltad_), vQ(vQ_),vB(vB_),Qb(Qb_){
    Qb = max(Qb_,alpha1/alpha2+100*Eps);
    //alpha1/alpha2 is the minimal Q that is possibile with this modelling
    mout <<  "Quality Factor set at BND: " << Qb << " Minimal in Modelling is: "
         <<alpha1/alpha2+100*Eps;
  }

};
vector<double> initializePMInftyVec(int size);
enum TAPERSTYLE {
  OLDTAPER = 0,
  COSTAPER = 1

};
class PMLConf {
    std::shared_ptr<CosTaperConfig> Cos = nullptr;
    vector<double> PMLPaddingLength;
    int taperstyle = OLDTAPER;
public:
    PMLConf(){}
    vector<int> pmlBndFlags;
    double distfactor = 1.0;
    void SetPMLPaddingLength(const vector<double> &PMLPL){
      PMLPaddingLength = PMLPL;
    }
    double GetPMLPaddingLength(int i) const {
      return PMLPaddingLength[i];
    }
    int getTaperStyle() const {
      return taperstyle;
    }
  CosTaperConfig& GetCosConfig() const{
      return *Cos;
    }
    void SetTaperStyle(int taperstyle_,double alpha1 = -1, double alpha2 = -1) {
        taperstyle = taperstyle_;
      if (taperstyle_ == COSTAPER){
        if(alpha1 ==-1 || alpha2 == -1) Exit("Specify alphas");
        double deltad = 0.3;
        double vQ = 0.85;
        double vB = 0.2;
        double Qb = 1.1;
        Config::Get("CosTaperEndCosPart",deltad);
        Config::Get("CosTaperRelativeValueAtEndOfCos",vQ);
        Config::Get("CosTaperRelativeVelocityAtBND",vB);
        Config::Get("CosTaperDampingAtBND",Qb);
        Cos = std::make_shared<CosTaperConfig>(alpha1,alpha2,deltad,vQ,vB,Qb);
      }
  }
};

using TransferPML = std::unordered_map<Point, std::pair<short, short>>;
using TransferToPMLBoundary = std::unordered_map<Point, std::vector<std::pair<Point, short>>>;
using PointToPoint = std::unordered_map<Point,Point>;
class PMLHandler {
    void addCornerPoints(const vector<vector<double>> &outerBoundaryPoints, vector<Point> &Cmid, int dim, int nrbounds);
    TransferPML PMLToNonPML;
    TransferPML NonPMLToPML;
    std::vector<double> dx;
    TransferToPMLBoundary NonPMLBoundaryToPML;
    int numL;
    double distfactor;
    std::unique_ptr<EleCacheContainer> eleCacheContainerPML;
    PMLConf ConfigPML;
    FWIDiscretizationStack FWIDiscStackPML;
    vector<double> MeshBordersPML;
    vector<double> ProcBordersPML;
    vector<double> MeshBorders;
    vector<double> ProcBorders;
    bool initialized = false;

  struct CommDataRed{
    int sendProc;
    int ownerCellProc;
    Point CellPoint;
  };
  struct CommData{
    int sendProc;
    int borderProc;
    Point BorderPoint;
    Point PMLPoint;
  };
    void findBndFlags(const Mesh &, const vector<double> &, vector<int> &, vector<int> &);

    void
    findCellNrAndPMLLength(const vector<double> &, vector<int> &, vector<int> &, vector<double> &, const vector<int> &,
                           int, std::vector<double> &);

    void initializePMLMesh(const Mesh &mesh, const vector<double> &MeshBordersI);


    void fillDiscStack(int deg, int numL, int paramComps);

    void initializeEleCache(const Vector &vec_, int numL);

public:
  PointToPoint PMLToBoundary;

    static void findMeshAndProcBorders(const Mesh &mesh, vector<double> &MeshBorders, vector<double> &ProcBorders);
    PMLHandler() {}

    const PMLConf &GetConfig() const { return ConfigPML;}

    int GetNumL() const { return numL; }
    void CreatePointsOutsideOfInnerGrid(const Mesh &VecMesh, vector<std::pair<Point, vector<Point>>> &Cmids);

    int isOnProc(const Point &P, const vector<double> &VectorOfBoxes);

    void InitializeBoundaryCellToTransfer(TransferToPMLBoundary &BondaryCellToTransfer,
                                          std::vector<std::pair<Point, std::vector<Point>>>
                                          &ExtrapolatedPoints, const vector<double> &Boundaries);

  void InitializeBoundaryCellToTransferNEW(TransferToPMLBoundary &BondaryCellToTransfer,
                                           vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                           const vector<double> &Boundaries,
                                           const Vector &meshPML);
    void InitializeCellToTransfer(TransferPML &CellToTransfer, const Mesh &, const vector<double> &Boundaries);
   void InitializeCellToTransferNEW(TransferPML &CellToTransfer, const Mesh &meshFrom, const Mesh& meshTo,
  const vector<double> &Boundaries);
    void Initialize(const Mesh &mesh, PMLConf Conf, int deg, int numL, int paramComps);

    const FWIDiscretizationStack &GetDiscStack() const;

    FWIDiscretizationStack &GetDiscStack();

    const EleCacheType &GetEleCache() const;


    double GetDx() const {
        for (int i = 0; i < dx.size() - 1; ++i) {
            if (dx[i] != dx[i + 1]) {
                THROW("find usages of this and adapt it to differing dx");
            }
        }
        return dx[0];
    }

    bool isPML() const;

    const Vector &GetWaveVector() const;

    const Vector &GetMatVector() const;

    const vector<double> &GetProcBoundariesPML() const;

    const vector<double> &GetProcBoundaries() const;

    const vector<double> &GetMeshBordersPML() const;

    const vector<double> &GetMeshBorders() const;

    const TransferPML &GetPMLToNonPML() const;

    const TransferPML &GetNonPMLToPML() const;

    const TransferToPMLBoundary &GetNonPMLBoundaryToPML() const;

    double GetDistFactor() const {
        return ConfigPML.distfactor;
    }

    ~PMLHandler();
};
class AAOPMLConfig {
  std::shared_ptr<CosTaperConfig> Cos = nullptr;
  int taperstyle = OLDTAPER;
public:
  RectangularConfig RectConf;
  double distfactor = 1.0;
  AAOPMLConfig(){
    Config::Get("AAOPMLDistFactor", distfactor);
    Config::Get("InnerLeftBnd", RectConf.LeftBnd);
    Config::Get("InnerRightBnd", RectConf.RightBnd);
    Config::Get("InnerBottomBnd", RectConf.BottomBnd);
    Config::Get("InnerTopBnd", RectConf.TopBnd);
    Config::Get("InnerDimX", RectConf.dimX);
    Config::Get("InnerDimY", RectConf.dimY);
    int bnd = 0;
    RectConf.bndTop = bnd;
    RectConf.bndBottom = bnd;
    RectConf.bndLeft = bnd;
    RectConf.bndRight = bnd;
    RectConf.name = "Rect";

  }
  CosTaperConfig& GetCosConfig() const{
    return *Cos;
  }
  void SetTaperStyle(int taperstyle_,double alpha1 = -1, double alpha2 = -1) {
    taperstyle = taperstyle_;
    if (taperstyle_ == COSTAPER){
      if(alpha1 ==-1 || alpha2 == -1) Exit("Specify alphas");
      double deltad = 0.3;
      double vQ = 0.85;
      double vB = 0.2;
      double Qb = 1.1;
      Config::Get("CosTaperEndCosPart",deltad);
      Config::Get("CosTaperRelativeValueAtEndOfCos",vQ);
      Config::Get("CosTaperRelativeVelocityAtBND",vB);
      Config::Get("CosTaperDampingAtBND",Qb);
      Cos = std::make_shared<CosTaperConfig>(alpha1,alpha2,deltad,vQ,vB,Qb);
    }
  }
  int getTaperStyle(){ return taperstyle;}
};
enum {
  LEFT =0,
  RIGHT = 1,
  DOWN= 2,
  UP =3,
};
class PMLHandlerAAO{
  double DistFactor = 1.0;
  std::vector<double> dx;
  double distanceFactor(double d) {
        if (DistFactor < Eps) return 1;
        if (DistFactor * d / dx[0] < 1 ) return 1;
        return DistFactor * d / dx[0];
    }
  vector<double> MeshBorders;
  vector<double> MeshBordersPML;
  vector<double> ProcBordersPML;
  LevelPair lev;
  TransferToPMLBoundary NonPMLBoundaryToPML;
  AAOPMLConfig Conf;
  AAOPMLConfig& GetConfig(){
    return Conf;
  }
  int numL;
  vector<double> PMLPaddingLength;
public:
  explicit PMLHandlerAAO(const AAOPMLConfig &Conf, const Vector &vecfull,int numL);

  int isOnProc(const Point &P, const vector<double> &VectorOfBoxes) {
    int dim = (P.SpaceDim() + P.TimeDim());
    for (int i = 0; i < PPM->size(); ++i) {
      bool cont = true;
      for (int j = 0; j < dim; ++j) {
        bool tmp = VectorOfBoxes[i * 2 * dim + dim * j] <= P[j];
        bool tmp2 = VectorOfBoxes[i * 2 * dim + dim * j + 1] >= P[j];
        if (!tmp || !tmp2) cont = false;
      }
      if (cont) {
        return i;
      }
    }
    return -1;
  }
  struct CommData{
    int sendProc;
    int borderProc;
    Point BorderPoint;
    Point PMLPoint;
  };
  void InitializeBoundaryCellToTransferAAO(TransferToPMLBoundary &BondaryCellToTransfer,
                                        vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                        const vector<double> &Boundaries,const Vector& meshfull);
  void SetPMLZero(Vector& Vec){
    for (cell c = Vec.cells(); c!=Vec.cells_end(); c++){
      if( c()[0] > MeshBorders[0] && c()[0] < MeshBorders[1]
      && c()[1] > MeshBorders[2] && c()[1] < MeshBorders[3]){
        continue;
      }
      int d = Vec.Dof(Vec.Id(c()));
      for (int i =0; i < d; ++i) {
        Vec(c(),i) = 0.0;
      }
    }
  }
  void CommunicateBoundaryToPML(Vector &Vec) {
    ExchangeBuffer exBuffer;
    for (const auto &it: NonPMLBoundaryToPML) {
      const auto &SendPoint = it.first;
      for (const auto &P: it.second) {
        short receiveProc = P.second;
        const auto &ReceivePoint = P.first;
        if (ReceivePoint == SendPoint) continue;
        exBuffer.Send(receiveProc) << ReceivePoint;
        bool found = Vec.find_cell(SendPoint) != Vec.cells_end();
        if (!found){
          THROW("Point" + to_string(SendPoint[0]) + " " + to_string(SendPoint[1])
          + " not found on proc " + to_string(PPM->Proc()));
        }
        int i = Vec.Id(SendPoint);
        for (int k = 0; k < Vec.Dof(i); ++k) {
          exBuffer.Send(receiveProc) << Vec(i, k);
        }
      }
    }
    Vec.CommunicateVector(exBuffer);
  }
  void increaseImpedance(Vector &Vec) {
    ExchangeBuffer exBuffer;
    for (const auto &it: NonPMLBoundaryToPML) {
      const auto &SendPoint = it.first;
      for (const auto &P: it.second) {
        short receiveProc = P.second;
        const auto &ReceivePoint = P.first;
        if (ReceivePoint == SendPoint) continue;
//        double d = -infty;
//        for (int i = 0; i < Vec.dim(); ++i) {
//          d = max(abs(ReceivePoint[i] - SendPoint[i]), d);
//        }
        exBuffer.Send(receiveProc) << ReceivePoint;
        exBuffer.Send(receiveProc) << SendPoint;

//        exBuffer.Send(receiveProc) << distanceFactor(d);
      }
    }
    communicateImpedanceFactors(exBuffer,Vec);
  }

  void communicateImpedanceFactors(ExchangeBuffer &exBuffer, Vector& Vec) {
    exBuffer.Communicate();
    for (short q = 0; q < PPM->Size(Vec.CommSplit()); ++q) {
      while (exBuffer.Receive(q).size() < exBuffer.ReceiveSize(q)) {
        Point ReceivePoint;
        exBuffer.Receive(q) >> ReceivePoint;
        int i = Vec.Id(ReceivePoint);
        Point SendPoint;
        exBuffer.Receive(q) >> SendPoint;
        applydistanceFactors(Vec,SendPoint,ReceivePoint,i);
      }
    }
    exBuffer.ClearBuffers();
  }
  double OldTaper(double dist){ // this is basically an exponential taper in the velocity
    if (DistFactor * dist / dx[0] < 1) return 1;
    return DistFactor * dist / dx[0];
  }
  std::pair<double,double> calculateRelativeDistance(double distX,double distY,
                                   vector<double> LayerSizes){
    double relDist = -infty;
    double Dist = -infty;
    if (abs(distX)>Eps) {
      int direc = distX < 0  ? LEFT : RIGHT;
      relDist = abs(distX) / LayerSizes[direc];
      Dist = LayerSizes[direc];
    }
    if (abs(distY) > Eps) {
      int direc = distY < 0  ? DOWN : UP;
      if (relDist <  abs(distY) / LayerSizes[direc]){
        relDist = abs(distY) / LayerSizes[direc];
        Dist = LayerSizes[direc];
      }
    }
    return {relDist,Dist};
  }
  void applydistanceFactors(Vector& Vec,
                            const Point &SendPoint, const Point &ReceivePoint, int i) {
    if (DistFactor < Eps) return;
    double distX = ReceivePoint[0] - SendPoint[0];
    double distY = ReceivePoint[1] - SendPoint[1];
    if (GetConfig().getTaperStyle() == OLDTAPER){
      double fac = OldTaper(max(abs(distY),abs(distX)));
      Vec(i,0) *= fac;
      for (int n =0; n <= numL; ++ n){
        Vec(i,1+n) /= fac;
      }
      return;
    }
    if(GetConfig().getTaperStyle() == COSTAPER){
      auto [relDist,Dist] = calculateRelativeDistance(distX,distY,PMLPaddingLength);
      double vQ = GetConfig().GetCosConfig().vQ;
      double vB = GetConfig().GetCosConfig().vB;
      double alpha1 = GetConfig().GetCosConfig().alpha1;
      double alpha2 = GetConfig().GetCosConfig().alpha2;
      double Qb = GetConfig().GetCosConfig().Qb;
      double Kappa0 = Vec(i, 1);
      double Kappa1 = Vec(i, 2);
      double taup = Kappa1 / Kappa0;
        double deltad = GetConfig().GetCosConfig().deltad;
        double C = relDist < deltad ? vQ + (1 - vQ) * cos(relDist / deltad * M_PI_2)
                                    : vQ + (vB - vQ) * (relDist - deltad) / (1 - deltad);
        if (numL == 0) {
          Vec(i, 0) /= C;
          Vec(i, 1) *= C;
        } else {
          double QM = (1.0 + taup * alpha1) / (taup * alpha2);
          double QT =
              pow(QM, 1.0 - relDist) * pow(Qb, relDist); // Exponential Taper from Q Value at
          double TauPML = 1.0 / (alpha2 * QT - alpha1);
          double quot1 = (1 + alpha1 * TauPML) / (1 + alpha1 * taup);
          double quot2 = (1.0 + numL * TauPML) / (1.0 + numL * taup);
          double RhoFac = sqrt(quot1 / quot2);
          double KappaFac = 1.0 / (sqrt(quot1 * quot2)); // these factors are supposed to ensure
          // the correct velocities as well as the correct impedance if the damping changes
          Vec(i, 0) *= RhoFac * 1.0 / C;
          Vec(i, 1) *= KappaFac * C;
          for (int n = 1; n <= numL; ++n) {
            Vec(i, 1 + n) = Vec(i, 1) * TauPML;
          }
        }

    }
    else Exit("Taper Not known");
  }
};
class PMLVectorFWI : public Vector {
protected:
    const PMLHandler &handler;


public:
    PMLVectorFWI(const PMLHandler &handler, const Vector &Vec) : Vector(0, Vec), handler(handler) {
        (*this).Clear();
    }

    static void transfer(const TransferPML &Transfer, const Vector &VecSend, Vector &VecRec);

    void PaddedToUnpadded(Vector &Vec);

    void UnpaddedToPadded(const Vector &Vec);

};



class PMLMatVectorFWI : public PMLVectorFWI {


  double OldTaper(double dist){ // this is basically an exponential taper in the velocity
    if (handler.GetDistFactor() * dist / handler.GetDx() < 1) return 1;
    return handler.GetDistFactor() * dist / handler.GetDx();
  }
    void applydistanceFactors(const Point &SendPoint,
                          const Point &ReceivePoint, int i,int j);

  double calculateRelativeDistance(double distX,double distY){
    double relDist = -infty;
    if (abs(distX)>Eps) {
      int direc = distX < 0  ? LEFT : RIGHT;
      relDist = abs(distX) / handler.GetConfig().GetPMLPaddingLength(direc);
    }
    if (abs(distY) > Eps) {
      int direc = distY < 0  ? DOWN : UP;
      relDist = max(relDist,abs(distY) / handler.GetConfig().GetPMLPaddingLength(direc));
    }
    return relDist;
  }
    void increaseImpedance();

    void communicateImpedanceFactors(ExchangeBuffer &exBuffer);

public:
    PMLMatVectorFWI(const PMLHandler &handler, const Vector &Unpadded, bool increaseImp);

    void CommunicateBoundaryToPML(const Vector &VecSend);
};

class PMLWaveVectorFWI : public PMLVectorFWI {
public:
    explicit PMLWaveVectorFWI(const PMLHandler &handler) : PMLVectorFWI(handler, handler.GetWaveVector()) {}
};

#endif //FWI_PMLHANDLER_HPP
