//
// Created by cr on 24.11.22.
//

#ifndef FWI_FWIACOUSTICINVERSIONPROB_HPP
#define FWI_FWIACOUSTICINVERSIONPROB_HPP

#include "Sources/FWISourceWavelet.hpp"
#include "Sources/WaypointSources.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "SourceInversion.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"

class FWIVAcousticInversionProb : public WaveProblemDGTSVA {
protected:
    double factor = 1.0;
  std::vector<std::unique_ptr<SourceWaveletBase>> srcSignals;

//  std::unique_ptr<SourceWaveletBase> TimeSource;
    std::unique_ptr<SourcesPositions> Sources;
    std::vector<double> tau_i;
    double wRef = 0.0;
    double alphaValue = 0.0;

    void calculateAlphaValue();

public:
    FWIVAcousticInversionProb( int numL, int dim)
            : WaveProblemDGTSVA(numL, dim) {
    }

    double ut(double, const Cell &, const Point &, int) const override;

    bool RHS(double) const override;

    void setInitialValueFlag(bool _flag) override;

    bool exact() const override;

    VectorField ForceV(double, const Cell &, const Point &, int) const override;

    int getNrPointSrc() const override;

    Point getPointSrcLoc(int nr) const override;

    int getShotCount() const override;

    void setShot(int _shot) override;

    int getShot() const override;

    double Kappa_i(const Cell &c, const Point &z, int i) const override;

    double Tau_i(const Cell &C, const Point &x, int i) const override;

    double alpha(const Cell &c, const Point &z) const override;

    double KappaRef(const Cell &C, const Point &z) const;
  double alpha1() const override;
  double alpha2() const override;
  double ForceP_i(double t, const Cell &c, const Point &z, int n, int) const override {
    return (n == 0) ? (*srcSignals[Sources->GetShot()])(t) * factor : 0.0;
  }
  virtual void SetTimeSource(std::unique_ptr<SourceWaveletBase> wav)  {
    for (int i = 0; i < Sources->GetShotCount(); ++i) {
      srcSignals[i] = std::make_unique<SourceWaveletBase>(*wav);
    }
  }
  virtual std::unique_ptr<SourceWaveletBase> MoveTimeSource()  {
    return std::move(srcSignals[Sources->GetShot()]);
  }
  SourceWaveletBase &GetTimeSourceRef()  {
    return *srcSignals[Sources->GetShot()];
  }
};
void SourceInversionProblem(const std::vector<FWIDataContainerRange> &obs,
                            const std::vector<FWIDataContainerRange> &obsFwd,
                            std::shared_ptr<WaveProblemDGTSVA>& prob,
                            const FWIObservationOperatorNEW& obsOp,
                            double eps, double Kappa, int j);

#endif //FWI_FWIACOUSTICINVERSIONPROB_HPP
