//
// Created by cr on 12.09.22.
//
#include "WaveProblemFromImages2D.hpp"

std::string makeNameStringPng(const std::string &fn_image) {
    return "FWI/models/" + fn_image + ".png";
}


WaveProblemFromImages2D::WaveProblemFromImages2D(const Point &minP, const Point &maxP, int _numL)
        : FWIVAcousticInversionProb( _numL, 2),
          LeftRight(minP[0], maxP[0]), BottomTop(minP[1], maxP[1]) {
    mout << "WaveProblemFromImages2d_RHS: mesh bounding box:"
         << "(" << LeftRight.first << ", " << LeftRight.second << ")" << " - "
         << "(" << BottomTop.first << ", " << BottomTop.second << ")" << endl;
    std::string param;
    Config::Get("parametrization", param);
    setParametrization(param);
    Config::Get("impedance_bg", impedance_bg);
    readImage("Rho", RhoImage, RhoInitialImage);
    readImage("Vp",VpImage,VpInitialImage);
    if (_numL > 0 ) readImage("Taup",TaupImage,TaupInitialImage);
}

void WaveProblemFromImages2D::readImage(const string &ParaName, std::shared_ptr<ParameterImage> &Image,
                                        std::shared_ptr<ParameterImage> &ImageInitial) {

    string img_fn;
    Config::Get("ModelImage" + ParaName, img_fn);
    Image = std::make_shared<ParameterImage>(ParaName, makeNameStringPng(img_fn), LeftRight, BottomTop);
    string imgIni_fn;
    Config::Get("ModelImage" + ParaName + "Initial", imgIni_fn);
    if (!imgIni_fn.empty())
        ImageInitial = std::make_shared<ParameterImage>(ParaName, makeNameStringPng(imgIni_fn), LeftRight,
                                                       BottomTop);
    else ImageInitial = Image;

}

double WaveProblemFromImages2D::Rho(const Cell &c, const Point &P) const {
  if (c()[0] < LeftRight.first || c()[0] > LeftRight.second || c()[1] < BottomTop.first || c()[1] > BottomTop.second) return 0.0;
    double rho = initial_value ? RhoInitialImage->ValueAtPoint(c()) : RhoImage->ValueAtPoint(c());
    return rho * impedance_bg;
}

double WaveProblemFromImages2D::V_p(const Cell &c, const Point &P) const {
  if (c()[0] < LeftRight.first || c()[0] > LeftRight.second || c()[1] < BottomTop.first || c()[1] > BottomTop.second) return 0.0;
    return initial_value ? VpInitialImage->ValueAtPoint(c()) : VpImage->ValueAtPoint(c());
}

double WaveProblemFromImages2D::Tau_p(const Cell &c, const Point &P) const {
  if (c()[0] < LeftRight.first || c()[0] > LeftRight.second || c()[1] < BottomTop.first || c()[1] > BottomTop.second) return 0.0;
    return initial_value ? TaupInitialImage->ValueAtPoint(c()) : TaupImage->ValueAtPoint(c());
}
