//
// Created by cr on 18.11.22.
//

#include "FWILinOpRed.hpp"

void FWILinOpRed::Solve(const Vector &m_test, Vectors &U_linearized) {
    FWITimeSolver TSolveR(U_m[0], PMLH, TConf.rkorder);
    TSolveR.Solve(getAssembler(m_test), U_linearized);
}

void FWILinOpRed::SolveAndMeasure(const Vector &m_test,
                                  const FWIObservationOperatorNEW &Obs,
                                  FWISeismogram &s_linearized, Vectors *U_linearized) {
    FWITimeSolver TSolveR(U_m[0], PMLH, TConf.rkorder);
    TSolveR.SolveAndMeasure(getAssembler(m_test), Obs, s_linearized, U_linearized);
}

FWILinOpRed::FWILinOpRed(const Vectors &U_m_,  const Vector &m_, const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                         const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                         const EleCacheType &eleCache_,
                         bool discrete_)
        : prob(prob_), U_m(U_m_), m(m_), PMLH(PMLH_), eleCache(eleCache_), TConf(TConf_), discrete(discrete_) {
    VAPlotter = std::make_shared<DGVAcousticPlotter>(prob->getNumL(), nullptr);
    auto Mat = std::make_unique<FWIDGVAcousticMaterialFromVector>(*prob, m, PMLH.get());
    matrixAssembler = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true);
}

std::shared_ptr<DGVAcousticAssembleCombined> FWILinOpRed::getAssembler(const Vector &m_test) {
    auto rhs = std::make_shared<DGVAcousticRHSAssemblerLinearized>(U_m, prob, eleCache, m, m_test, TConf.dt,
                                                                   discrete, PMLH.get());
    auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssembler, rhs,
                                                                          TConf.t0,
                                                                          TConf.T,
                                                                          TConf.dt);
    assembleCombined->SetEleCache(PMLH->isPML() ? &PMLH->GetEleCache() : &eleCache);
    return std::move(assembleCombined);
}

