//
// Created by cr on 18.11.22.
//

#ifndef FWI_FWILINOPRED_HPP
#define FWI_FWILINOPRED_HPP

#include "Vector.hpp"
#include "FWITimeSolver.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "DGVAcousticMatrixAssembler.hpp"
#include "FWIDGVAcousticMaterialFromVector.hpp"
#include "Assembler/FWIRhsAssemblerLinearized.hpp"
#include "NormalizedHelpers.hpp"
//using SeisNormPair = std::pair<const FWISeismogram&, const std::unordered_map<Point,double>&>;



class FWILinOpRed {
protected:
    const std::shared_ptr<WaveProblemDGTSVA> &prob;
    const Vectors &U_m;
    const Vector &m;
    const std::shared_ptr<PMLHandler> &PMLH;
    std::shared_ptr<DGVAcousticMatrixAssembler> matrixAssembler = nullptr;
    std::shared_ptr<DGVAcousticPlotter> VAPlotter;
    const EleCacheType &eleCache;
    const TimeConf &TConf;
    bool discrete = false;
    std::shared_ptr<DGVAcousticAssembleCombined> getAssembler(const Vector &m_test);

public:
    FWILinOpRed(const Vectors &U_m_, const Vector &m_, const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_, const EleCacheType &eleCache_,
                bool discrete);

    virtual void Solve(const Vector &m_test, Vectors &U_linearized);

    virtual void SolveAndMeasure(const Vector &m_test, const FWIObservationOperatorNEW &Obs,
                         FWISeismogram &s_linearized, Vectors *U_linearized);
};
class FWILinOpRedNormalized : public FWILinOpRed {
  const FWISeismogram &sMNormed;
  const std::unordered_map<Point,double> &sMNorms;
public:
  FWILinOpRedNormalized(const Vectors &U_m_, const FWISeismogram& sMNormed_,
                        const std::unordered_map<Point,double> &sMNorms_, const Vector &m_,
                        const std::shared_ptr<WaveProblemDGTSVA> &prob_,
                        const std::shared_ptr<PMLHandler> &PMLH_, const TimeConf &TConf_,
                        const EleCacheType &eleCache_,bool discrete)
                        : FWILinOpRed(U_m_,m_,prob_,PMLH_,TConf_,eleCache_,discrete),
                               sMNormed(sMNormed_), sMNorms(sMNorms_)
                        {}

  void Solve(const Vector &m_test, Vectors &U_linearized){ Exit("Makes no sense here")}

  void SolveAndMeasure(const Vector &m_test, const FWIObservationOperatorNEW &Obs,
                       FWISeismogram &s_linearized, Vectors *U_linearized){
    FWITimeSolver TSolveR(U_m[0], PMLH, TConf.rkorder);
    TSolveR.SolveAndMeasure(getAssembler(m_test), Obs, s_linearized, U_linearized);
    PrepareLinearizedAndAdjointNormalized(sMNormed,sMNorms,s_linearized);
  }
};

#endif //FWI_FWILINOPRED_HPP
