#ifndef _RegularisationStrategy_H_
#define _RegularisationStrategy_H_
#include <vector>
#include <ostream>
class RegularisationStrategy { //
protected:
  double gamma = -1.0;
  double thetaMax = -1.0;
    unsigned int allowedInnerIters;
    std::vector<double> thetas;
  std::vector<double> resNorms;
  std::vector<unsigned int> innerIters;
  std::vector<unsigned int> maxInnerIters;
public:
  RegularisationStrategy(double _gamma, double _thetaMax, int _allowedInnerIters) :
      gamma(_gamma), thetaMax(_thetaMax), allowedInnerIters(_allowedInnerIters) {}

  virtual ~RegularisationStrategy() = default;

  RegularisationStrategy() = default;

//    RegularisationStrategy(const RegularisationStrategy& ) = delete;
  void addResNorm(double res) { resNorms.push_back(res); }

  auto getResNorm(int iter) -> double const { return resNorms[iter]; }

  auto getResNormVector(int iter) -> std::vector<double> & { return resNorms; }

  auto getTheta(int iter) -> double const { return thetas[iter]; }

  void addTheta(double theta) { thetas.push_back(theta); }

  void addInnerIter(unsigned int innerIter) { innerIters.push_back(innerIter); }

  virtual auto calcTheta() -> double const = 0;
    virtual auto calcMaxInner() -> unsigned int const = 0;
    virtual auto calcInnerStopp() -> double const = 0;
    auto getInnerIter(int iter) const {
        return innerIters[iter];
    }
};

class RegularisationStrategyWinkler : public RegularisationStrategy {
public:
  ~RegularisationStrategyWinkler() = default;

  RegularisationStrategyWinkler() = default;

  RegularisationStrategyWinkler(double _gamma, double _thetaMax, int _allowedInnerIters) :
      RegularisationStrategy(_gamma, _thetaMax, _allowedInnerIters) {}

  auto calcInnerStopp() -> double const override {
    return thetas[thetas.size() - 1] * resNorms[thetas.size() - 1];
  }

  auto calcTheta() -> double const override {
    if (thetas.size() == 0) {
      thetas.push_back(thetaMax);
      return thetaMax;
    } else if (thetas.size() == 1) {
      thetas.push_back(resNorms[1] / resNorms[0]);
            return resNorms[1]/resNorms[0];
        }
        else {
      unsigned int iters_m2 = innerIters[innerIters.size() - 2];
      unsigned int iters_m1 = innerIters[innerIters.size() - 1];
//          MOUT(iters_m2)
//          MOUT(iters_m1)

      double increaseFactor = static_cast<double>(static_cast<double>(iters_m2) /
                                                  static_cast<double>(iters_m1));
      bool max = false;
      if (iters_m1 == allowedInnerIters && iters_m2 == allowedInnerIters) {
        increaseFactor = 0.5;
        max = true;
      }
      double thetaTilda = (iters_m1 > iters_m2 || max) ?
                          1.0 - increaseFactor * (1.0 - thetas[thetas.size() - 1]) : gamma * thetas[
              thetas.size() - 1];
//          MOUT(increaseFactor);
//          MOUT(thetas[thetas.size()-1]);
//
//          MOUT(increaseFactor * ( 1.0 - thetas[thetas.size()-1]));
      thetas.push_back(std::min(thetaMax, thetaTilda));
      return std::min(thetaMax, thetaTilda);
    }}
    auto calcMaxInner() -> unsigned int const override{
        if (innerIters.size()==0){maxInnerIters.push_back(1);
            return 1u;
        }
        else if (innerIters.size()==1) {maxInnerIters.push_back(2);
            return min(allowedInnerIters, 2u);
        }
        else {
            unsigned int iters_m2 = innerIters[innerIters.size()-2];
            unsigned int iters_m1 = innerIters[innerIters.size()-1];
            maxInnerIters.push_back((iters_m1 + iters_m2));
            return min(allowedInnerIters, (iters_m1 + iters_m2));
    }
  }
};
#endif
