//
// Created by cr on 23.08.22.
//

#include "DGVAcousticAssembleCombined.hpp"


const char *DGVAcousticAssembleCombined::Name() const {
    return nullptr;
}

DGVAcousticAssembleCombined::~DGVAcousticAssembleCombined() {
    if (clearCache) {
        for (auto it = EleCache->begin(); it != EleCache->end(); ++it) {
            delete it->second;
        }
    }
}

void DGVAcousticAssembleCombined::Initialize(Vector &u) const {
    rhsAssembler->Initialize(u);
}

void DGVAcousticAssembleCombined::RHS(double t, Vector &rhs) const {
    rhsAssembler->RHS(t, rhs);
}

void DGVAcousticAssembleCombined::MassMatrix(Matrix &massMatrix) const {
    assembler->MassMatrix(massMatrix);
    massMatrix.CreateSparse();
    massMatrix.setCollectAfterProd(!useCollectTweak);
}

void DGVAcousticAssembleCombined::SystemMatrix(Matrix &systemMatrix) const {
    assembler->SystemMatrix(systemMatrix);
    systemMatrix.CreateSparse();
    systemMatrix.setCollectAfterProd(!useCollectTweak);
}

void DGVAcousticAssembleCombined::FinishTimeStep(const Vector &u) {
    PlotIteration(u);
    PrintIteration(u);
}

void DGVAcousticAssembleCombined::SetInitialValue(Vector &u) {
    if (u0) u = *u0;
    else u.Clear();
//    InterpolatedExactSolution(u);
}

void DGVAcousticAssembleCombined::PrintInfo() const {
    ILinearTimeAssemble::PrintInfo();
}

void DGVAcousticAssembleCombined::InterpolatedExactSolution(Vector &u) const {
//    for (cell c = u.cells(); c != u.cells_end(); ++c) {
//        Point z_c = c();
//        row r = u.find_row(z_c);
//        norm(u);
//        const DGAcousticElement &elem = *EleCache->at(c());
//        for (int i = 0; i < elem.u_dimension(); ++i) {
//            for (int j = 0; j < elem.shape_size(); ++j) {
//                Point z = (1 - 1e-10) * elem.NodalPoint(j) + 1e-10 * z_c;
//                u(r)[i * elem.shape_size() + j] = assembler->GetProblem().ut(Time(), c, z, i);
//            }
//        }
//    }
}

AcousticErrorsDG DGVAcousticAssembleCombined::ComputeErrors(const Vector &u) const {
    AcousticErrorsDG errors;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const DGAcousticElement &elem = *EleCache->at(c());
        int p_dim = elem.p_dimension();
        for (int q = 0; q < elem.nQ(); ++q) {
          double w = elem.QWeight(q);
          Point z = elem.QPoint(q);

          VectorField exactV = zero;
          std::vector<double> exactP(p_dim, 0.0);
          Exit("Find a way to work around this");
//            for (int i = 0; i < assembler->GetProblem().Dim(); ++i)
//                exactV[i] = assembler->GetProblem().ut(Time(), c, z, i);
//            for (int n = 0; n < p_dim; ++n)
//                exactP[n] = assembler->GetProblem().ut(Time(), c, z, assembler->GetProblem().Dim() + n);

          VectorField V = elem.Velocity(q, u);
          V -= exactV;
          errors.l2 += w * V * V;
          errors.l1 += w * norm(V);
          errors.inf = max(errors.inf, norm(V));
          for (int n = 0; n < p_dim; ++n) {
            double P_i = elem.Pressure_i(q, u, n) - exactP[n];
            errors.l1 += w * abs(P_i);
            errors.l2 += w * P_i * P_i;
                errors.inf = max(errors.inf, abs(P_i));
            }
        }
    }
    errors.l1 = PPM->SumOnCommSplit(errors.l1, u.GetMesh().CommSplit());
    errors.l2 = sqrt(PPM->SumOnCommSplit(errors.l2, u.GetMesh().CommSplit()));
    errors.inf = PPM->Max(errors.inf, u.GetMesh().CommSplit());
    return errors;
}

double DGVAcousticAssembleCombined::Energy(const Vector &u) const {
    Scalar e = 0;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const Cell &C = *c;
        const DGAcousticElement &elem = *EleCache->at(c());
        e += Energy_local(u, elem, c);
    }
    return 0.5 * PPM->Sum(e);
}

double
DGVAcousticAssembleCombined::Energy_local(const Vector &u, const DGAcousticElement &elem, const cell &c) const {
    Scalar e = 0;
    for (int q = 0; q < elem.nQ(); ++q) {
        double w = elem.QWeight(q);
        Point z = elem.QPoint(q);
        int p_dim = elem.p_dimension();
        double rho = assembler->GetMaterial().Rho(*c, z);
        VectorField V = elem.Velocity(q, u);
        e += w * rho * V * V;
        for (int n = 0; n < p_dim; ++n) {
            double P_i = elem.Pressure_i(q, u, n);
          double invKappa = assembler->GetMaterial().invKappa_i(*c, z, n);
            e += w * invKappa * P_i * P_i;
        }
    }
    return e;
}

MinMaxPDG DGVAcousticAssembleCombined::MinMaxPressure(const Vector &u) const {
    MinMaxPDG minMaxP = {0.0, 0.0};
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const DGAcousticElement &elem = *EleCache->at(c());
        double P = 0;
        double PInt = 0;
        for (int q = 0; q < elem.nQ(); ++q) {
            double w = elem.QWeight(q);
            P += w * elem.Pressure(q, u);
        }
        double area = elem.Area();
        P *= (1.0 / area);
        PInt *= (1.0 / area);

        if (minMaxP.second < P) minMaxP.second = P;
        if (minMaxP.first > P) minMaxP.first = P;
    }
    minMaxP.second = PPM->Max(minMaxP.second, u.GetMesh().CommSplit());
    minMaxP.first = PPM->Min(minMaxP.first, u.GetMesh().CommSplit());
    if (abs(minMaxP.second) < 1e-10) minMaxP.second = 0;
    if (abs(minMaxP.first) < 1e-10) minMaxP.first = 0;
    return minMaxP;
}

AcousticNormsDG DGVAcousticAssembleCombined::ComputeNorms(const Vector &u) const {
    AcousticNormsDG norms;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const DGAcousticElement &elem = *EleCache->at(c());
        for (int q = 0; q < elem.nQ(); ++q) {
            double w = elem.QWeight(q);
            Point z = elem.QPoint(q);
          int p_dim = elem.p_dimension();
          double rho = assembler->GetMaterial().Rho(*c, z);
          VectorField V = elem.Velocity(q, u);
            norms.l2 += w * V * V;
            norms.l1 += w * norm(V);
            norms.energy += w * rho * V * V;
            for (int n = 0; n < p_dim; ++n) {
                double P_i = elem.Pressure_i(q, u, n);
                norms.l2 += w * P_i * P_i;
                norms.l1 += w * abs(P_i);
              double invKappa = assembler->GetMaterial().invKappa_i(*c, z, n);
                norms.energy += w * invKappa * P_i * P_i;
            }
        }
    }
    norms.l2 = sqrt(PPM->SumOnCommSplit(norms.l2, u.GetMesh().CommSplit()));
    norms.l1 = sqrt(PPM->SumOnCommSplit(norms.l1, u.GetMesh().CommSplit()));
    norms.energy = sqrt(PPM->SumOnCommSplit(norms.energy, u.GetMesh().CommSplit()));
    return norms;
}

void DGVAcousticAssembleCombined::PrintIteration(const Vector &u) const {
  if (!plotter->PlotTStep(Step())) return;
  if (false) { // find a fix
    Vector u_int(u);
    InterpolatedExactSolution(u_int);
    AcousticErrorsDG err_h = ComputeErrors(u);
    AcousticErrorsDG err_int = ComputeErrors(u_int);

    mout.PrintIteration(
            verbose,
            PrintIterEntry("n", Step(), 5, 1),
            PrintIterEntry("t", Time(), 13, 1),
            PrintIterEntry("Energy", Energy(u), 13, 1),
                PrintIterEntry("IErr_L1", err_int.l1, 13, 1),
                PrintIterEntry("IErr_L2", err_int.l2, 13, 1),
                PrintIterEntry("IErr_infty", err_int.inf, 13, 1),
                PrintIterEntry("Err_L1", err_h.l1, 13, 1),
                PrintIterEntry("Err_L2", err_h.l2, 13, 1),
                PrintIterEntry("Err_infty", err_h.inf, 13, 1)
        );
    } else {
        MinMaxPDG minMaxP = MinMaxPressure(u);
        AcousticNormsDG norms = ComputeNorms(u);
        mout.PrintIteration(
                verbose,
                PrintIterEntry("n", Step(), 5, 1),
                PrintIterEntry("t", Time(), 13, 1),
                PrintIterEntry("Energy", Energy(u), 13, 1),
                PrintIterEntry("MaxP", minMaxP.second, 13, 1),
                PrintIterEntry("MinP", minMaxP.first, 13, 1),
                PrintIterEntry("L1", norms.l1, 13, 1),
                PrintIterEntry("L2", norms.l2, 13, 1)
        );
    }
}

void DGVAcousticAssembleCombined::PlotIteration(const Vector &u) const {
  if (!plotter->PlotTStep(Step())) return;
  if (Step() == 0) {
    plotter->plotMatVec(u, assembler->GetMaterial());
    plotter->plotProcs(u);
  }
  plotter->plotWaveVector(u, Step());
}

void DGVAcousticAssembleCombined::SetEleCache(const std::unordered_map<Point, DGAcousticElement *> *eleCache) {
    EleCache = eleCache;
    plotter->SetElementCache(eleCache);
    assembler->SetElementCache(eleCache);
    rhsAssembler->SetElementCache(eleCache);
}
