//
// Created by cr on 14.07.21.
//

#include "DGVAcousticPlotter.hpp"

void DGVAcousticPlotter::plotMatVec(const Vector &u, const DGVAcousticMaterial &Mat) const {
    auto LvP = u.GetMesh().Level();
    LagrangeDiscretization disc(u.GetDisc().GetMeshes(), 0);
    Vector Rho(0.0, disc, LvP);
    Vector Procs(0.0, disc, LvP);
    Vector Kappa(0.0, disc, LvP);
    Vector Imp(0.0, disc, LvP);
    Vector Vp(0.0, disc, LvP);
    Vector Qp(0.0, disc, LvP);
    Vector Taup(0.0, disc, LvP);

  std::vector<Vector> KappaLs;
    for (size_t i = 0; i < Mat.GetNumDamping() + 1; ++i) {
        KappaLs.emplace_back(Vector(0.0, disc, LvP));
    }
    std::vector<double> rho = {infty, -infty};
    std::vector<double> kappa = {infty, -infty};
    std::vector<double> v = {infty, -infty};
    for (cell c = Rho.cells(); c != Rho.cells_end(); ++c) {
        Procs(c(), 0) = PPM->proc();
        Rho(c(), 0) = Mat.Rho(*c, c());
      Kappa(c(), 0) = Mat.Kappa(*c, c());
      Imp(c(),0) = sqrt(Rho(c(),0) * Kappa(c(),0));
      if(Mat.GetNumDamping() > 0){
        Taup(c(),0) = Mat.Kappa_i(*c, c(), 1)/Mat.Kappa_i(*c, c(), 0);
        Vp(c(),0) = sqrt(Mat.Kappa_i(*c,c(),0)*(1.0+Mat.alpha(*c,c())*Taup(c(),0))/Rho(c(),0));
        double val1 = (1.0 + Mat.alpha1() * Taup(c(),0));
        double val2 = Mat.alpha2() * Taup(c(),0);
        Qp(c(),0) = val1 / val2;
      }else{
        Vp(c(),0) = sqrt(Kappa(c(),0)/Rho(c(),0));
      }
      for (int n = 0; n < KappaLs.size(); ++n) {
        KappaLs[n](c(), 0) = Mat.Kappa_i(*c, c(), n);
      }
      rho[0] = min(rho[0], Rho(c(), 0));
      rho[1] = max(rho[1], Rho(c(), 0));
      kappa[0] = min(kappa[0], Kappa(c(), 0));
      kappa[1] = max(kappa[1], Kappa(c(), 0));
      double vTmp = Vp(c(),0);
      v[0] = min(v[0], vTmp);
      v[1] = max(v[1], vTmp);
    }
    rho[0] = PPM->Min(rho[0]);
    rho[1] = PPM->Max(rho[1]);
    kappa[0] = PPM->Min(kappa[0]);
    kappa[1] = PPM->Max(kappa[1]);
    v[0] = PPM->Min(v[0]);
    v[1] = PPM->Max(v[1]);

    string lvStr = "_Lv" + std::to_string(LvP.space);

    if (opt.plot_rho) {
        auto &Plot = mpp::plot("Rho" + lvStr);
        Plot.AddData("Rho", Rho);
        Plot.PlotFile();
        Plotting::Instance().RemovePlot("Rho" + lvStr);

    }
    if (opt.plot_kappa_0) {
        auto &Plot = mpp::plot("KappaSum" + lvStr);
        Plot.AddData("KappaSum", Kappa);
        Plot.AddData("Impedance", Imp);
        if(numL>0)Plot.AddData("Taup", Taup);
        if(numL>0)Plot.AddData("Qp", Qp);
        Plot.AddData("Vp", Vp);
        Plot.PlotFile();
        Plotting::Instance().RemovePlot("KappaSum" + lvStr);


    }
    if (opt.plot_kappa_l) {
        for (int l = 1; l < KappaLs.size(); ++l) {
            auto &Plot = mpp::plot("KappaL" + lvStr);
            Plot.AddData("KappaL", KappaLs[l]);
//            mpp::plot(("kappa_" + std::to_string(l) + lvStr).c_str()) << KappaLs[l] << mpp::save_plot(
//                    ("kappa_" + std::to_string(l) + lvStr).c_str());
            Plot.PlotFile();
            Plotting::Instance().RemovePlot("KappaL" + lvStr);

        }
    }
    vout(1) << " ----- " << "Rho in   [" << rho[0] << "," << rho[1] << "]"
            << " ----- "
            << "Kappa in [" << kappa[0] << ","
            << kappa[1] << "]" << " ----- "
          << "Vp in [" << v[0] << "," << v[1] << "]" << endl;
  if ((kappa[0] <= 0) || (rho[0] <= 0)) Exit("not admissible material");
}

void DGVAcousticPlotter::plotProcs(const Vector &u) const {
    auto LvP = u.GetMesh().Level();
    LagrangeDiscretization disc(u.GetDisc().GetMeshes(), 0);
    Vector Procs(0.0, disc, LvP);
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        Procs(c(), 0) = PPM->proc();
    }
    string lvStr = "_Lv" + std::to_string(LvP.space);

    mpp::plot("Procs" + lvStr) << Procs << mpp::save_plot("Procs" + lvStr);
    Plotting::Instance().RemovePlot("Procs" + lvStr);
}

void DGVAcousticPlotter::plotWaveVector(const Vector &u, int step,const char *name) const {
    auto LvP = u.GetMesh().Level();
    LagrangeDiscretization disc(u.GetDisc().GetMeshes(), 0, 1);
    LagrangeDiscretization disc2(u.GetDisc().GetMeshes(), 0, u.GetMesh().dim());

    Vector VComponents(0.0, disc2, LvP);
    std::vector<Vector> PComponents;
    for (size_t i = 0; i < numL + 1; ++i) {
        PComponents.emplace_back(Vector(Vector(0.0, disc, LvP)));
    }
    Vector PSumComponent(Vector(0.0, disc, LvP));
    Vector PIntComponent(Vector(0.0, disc, LvP));

    double tmp = 0;
    const Mesh &Mu = u.GetMesh();

    for (cell c = PSumComponent.cells(); c != PSumComponent.cells_end(); ++c) {
        const Cell &C = *c;
        DGAcousticElement *elem = (EleCache == nullptr) ? new DGAcousticElement(u, C, numL)
                                                        : EleCache->find(c())->second;
        double Area = elem->Area();
        for (int q = 0; q < elem->nQ(); ++q) {
          double w = elem->QWeight(q);

          for (size_t i = 0; i < PComponents.size(); ++i) {
              PComponents[i](c(), 0) += w * elem->Pressure_i(q, u, i) * (1.0 / Area);
          }
          PSumComponent(c(), 0) += w * elem->Pressure(q, u) * (1.0 / Area);
          if (EleCache == nullptr) delete elem;
      }
  }
    if (opt.plotV) {
        for (cell c = VComponents.cells(); c != VComponents.cells_end(); ++c) {
            const Cell &C = *c;
            DGAcousticElement *elem = (EleCache == nullptr) ? new DGAcousticElement(u, C, numL)
                                                            : EleCache->find(c())->second;
            VectorField V = zero;
            double Area = elem->Area();
            for (int q = 0; q < elem->nQ(); ++q) {
                double w = elem->QWeight(q);
                V += w * elem->Velocity(q, u) * (1.0 / Area);

            }
            for (int i = 0; i < VComponents.GetMesh().dim(); ++i) {
                VComponents(c(), i) = V[i];
            }
            if (EleCache == nullptr) delete elem;
        }
    }
    string lvStr = "_Lv" + std::to_string(LvP.space);
    if (name == nullptr) {
        char filename[127];
        if (opt.plotV) {
            auto &Plot = mpp::plot(NumberName(("V" + lvStr).c_str(), filename, step));
            Plot.AddData("V", VComponents);
            Plot.PlotFile();
            Plotting::Instance().RemovePlot(NumberName(("V" + lvStr).c_str(), filename, step));

        }
        auto &Plot = mpp::plot(NumberName(("PSum" + lvStr).c_str(), filename, step));
        Plot.AddData("PSum", PSumComponent);
        Plot.PlotFile();
        Plotting::Instance().RemovePlot(NumberName(("PSum" + lvStr).c_str(), filename, step));

//        mpp::plot("PSum" + lvStr) << PSumComponent
//                                  << mpp::save_plot(NumberName(("PSum" + lvStr).c_str(), filename, step));
    } else {
        std::string Name(name);
        std::string PSumStr;
        PSumStr = "PSum";
        std::string filenamePSum = nr_str((Name + PSumStr + lvStr).c_str(), step);
        auto &Plot = mpp::plot(filenamePSum);
        Plot.AddData("PSum", PSumComponent);
        Plot.PlotFile();
        Plotting::Instance().RemovePlot(filenamePSum);
//        mpp::plot("PSum" + lvStr) << PSumComponent << mpp::save_plot(filenamePSum);
        if (opt.plotV) {
            std::string VStr;
            VStr = "V";
            std::string filenameV = nr_str((Name + VStr + lvStr).c_str(), step);
            auto &PlotV = mpp::plot(filenameV);
            PlotV.AddData("V", VComponents);
            PlotV.PlotFile();
            Plotting::Instance().RemovePlot(filenameV);
        }
    }
}

DGVAcousticPlotter::DGVAcousticPlotter(int numL, const std::unordered_map<Point, DGAcousticElement *> *EleCache)
        : EleCache(EleCache), numL(numL) {
    Config::Get("plot_tStep", opt.plotTStep);
    Config::Get("plotV", opt.plotV);
    Config::Get("plot_rho", opt.plot_rho);
    Config::Get("plot_kappa_0", opt.plot_kappa_0);
    Config::Get("plot_kappa_l", opt.plot_kappa_l);
    Config::Get("plot_velocity", opt.plot_velocity);

}
