import numpy as np
from seismogram_lib import FWISeismogram, tracewisedifference,read_seismograms, toNpz
from acoustic_exact_solution_2D import create_pressure_seismogram, ricker
import multiprocessing as mp
import os
if __name__ == "__main__":
    import sys
    import argparse
    parser = argparse.ArgumentParser(description='AcousticSol')
    parser.add_argument("--SrcLoc", type=float,nargs=2)
    parser.add_argument("--wavespeed", type=float, default=3500)
    parser.add_argument("-f", type=float, default= 25.0)
    parser.add_argument("-s", type=str, default="",nargs='+')
    parser.add_argument("--SrcFactor", type=float, default=1.0)
    parser.add_argument("--ExactSeismoName", type=str, default="")
    parser.add_argument("-p", type=str, default="")
    parser.add_argument("--clean", action="store_true")
    args = parser.parse_args()
    if args.s =="":
        exit("Set SeismogramName!!")
    conf= {}
    conf["c"] = args.wavespeed
    conf["rho"] = 2000.0
    conf["f"] = args.f
    conf["source_pos"] = np.array([args.SrcLoc[0],args.SrcLoc[1]]),
    conf["source_t_offset"] = 1.5/conf["f"]
    conf["source_func"] = ricker
    conf["source_time_interval"] = [-2./conf["f"],2./conf["f"]]
    conf["source_factor"] = args.SrcFactor
    s,(d_min,d_max) = read_seismograms(args.s, r_step=1,component=0)
    s_sol = [create_pressure_seismogram(s[0], conf, 2)]
    s_sol.extend(s)
    tracewisedifference(s_sol,args.p)
    nrproc = 32
    if args.clean is True:
        with mp.Pool(nrproc) as pool:
            pool.starmap(toNpz, [(args.p,s) for s in s_sol])
        for name in args.s:
            os.remove(name)
